import datetime
import os
import re
import sys
import tempfile
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple

import toml

from strictdoc import __version__
from strictdoc.backend.reqif.sdoc_reqif_fields import ReqIFProfile
from strictdoc.backend.sdoc.constants import SDocMarkup
from strictdoc.cli.cli_arg_parser import (
    ExportCommandConfig,
    ServerCommandConfig,
)
from strictdoc.core.environment import SDocRuntimeEnvironment
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.file_modification_time import get_file_modification_time
from strictdoc.helpers.md5 import get_md5
from strictdoc.helpers.net import is_valid_host
from strictdoc.helpers.path_filter import validate_mask


def parse_relation_tuple(column_name: str) -> Optional[Tuple[str, str]]:
    match_result = re.search(
        r"^((Parent|Child|File)?)(\[(.{1,32})])?$", column_name
    )
    if match_result is None:
        return None
    return match_result.group(1), match_result.group(4)


class ProjectFeature(str, Enum):
    # Stable features.
    TABLE_SCREEN = "TABLE_SCREEN"
    TRACEABILITY_SCREEN = "TRACEABILITY_SCREEN"
    DEEP_TRACEABILITY_SCREEN = "DEEP_TRACEABILITY_SCREEN"

    MATHJAX = "MATHJAX"

    # Experimental features.
    SEARCH = "SEARCH"
    HTML2PDF = "HTML2PDF"
    REQIF = "REQIF"
    DIFF = "DIFF"
    PROJECT_STATISTICS_SCREEN = "PROJECT_STATISTICS_SCREEN"
    STANDALONE_DOCUMENT_SCREEN = "STANDALONE_DOCUMENT_SCREEN"
    TRACEABILITY_MATRIX_SCREEN = "TRACEABILITY_MATRIX_SCREEN"
    REQUIREMENT_TO_SOURCE_TRACEABILITY = "REQUIREMENT_TO_SOURCE_TRACEABILITY"
    SOURCE_FILE_LANGUAGE_PARSERS = "SOURCE_FILE_LANGUAGE_PARSERS"

    MERMAID = "MERMAID"
    RAPIDOC = "RAPIDOC"
    NESTOR = "NESTOR"

    ALL_FEATURES = "ALL_FEATURES"

    @staticmethod
    def all() -> List[str]:  # noqa: A003
        return list(map(lambda c: c.value, ProjectFeature))


@auto_described
class ProjectConfig:
    DEFAULT_PROJECT_TITLE = "Untitled Project"
    DEFAULT_DIR_FOR_SDOC_ASSETS = "_static"
    DEFAULT_DIR_FOR_SDOC_CACHE = "output/_cache"

    DEFAULT_FEATURES: List[str] = [
        ProjectFeature.TABLE_SCREEN,
        ProjectFeature.TRACEABILITY_SCREEN,
        ProjectFeature.DEEP_TRACEABILITY_SCREEN,
        ProjectFeature.SEARCH,
    ]
    DEFAULT_SERVER_HOST = "127.0.0.1"
    DEFAULT_SERVER_PORT = 5111
    DEFAULT_BUNDLE_DOCUMENT_VERSION = "@GIT_VERSION (Git branch: @GIT_BRANCH)"
    DEFAULT_BUNDLE_DOCUMENT_COMMIT_DATE = "@GIT_COMMIT_DATETIME"
    DEFAULT_SECTION_BEHAVIOR = "[SECTION]"

    def __init__(
        self,
        environment: SDocRuntimeEnvironment,
        project_title: str,
        dir_for_sdoc_assets: str,
        dir_for_sdoc_cache: str,
        project_features: List[str],
        server_host: str,
        server_port: int,
        include_doc_paths: List[str],
        exclude_doc_paths: List[str],
        source_root_path: Optional[str],
        include_source_paths: List[str],
        exclude_source_paths: List[str],
        test_report_root_dict: Dict[str, str],
        source_nodes: List[Dict[str, str]],
        html2pdf_template: Optional[str],
        bundle_document_version: Optional[str],
        bundle_document_date: Optional[str],
        traceability_matrix_relation_columns: Optional[
            List[Tuple[str, Optional[str]]]
        ],
        reqif_profile: str,
        reqif_multiline_is_xhtml: bool,
        reqif_enable_mid: bool,
        reqif_import_markup: Optional[str],
        config_last_update: Optional[datetime.datetime],
        chromedriver: Optional[str],
        section_behavior: Optional[str],
        statistics_generator: Optional[str],
    ) -> None:
        assert isinstance(environment, SDocRuntimeEnvironment)
        if source_root_path is not None:
            assert os.path.isdir(source_root_path), source_root_path
            assert os.path.isabs(source_root_path), source_root_path

        self.environment: SDocRuntimeEnvironment = environment

        # Settings obtained from the strictdoc.toml config file.
        self.project_title: str = project_title
        self.dir_for_sdoc_assets: str = dir_for_sdoc_assets

        if env_cache_dir := os.environ.get("STRICTDOC_CACHE_DIR"):
            # The only use case for STRICTDOC_CACHE_DIR is to make the cache
            # local to an itest folder.
            assert env_cache_dir == "Output/_cache", env_cache_dir
            dir_for_sdoc_cache = env_cache_dir
        elif dir_for_sdoc_cache == "$TMPDIR":
            dir_for_sdoc_cache = os.path.join(
                tempfile.gettempdir(),
                "strictdoc_cache",
                get_md5(os.getcwd()),
            )

        # Adding a __version__ part to the cache directory improves traceability
        # by indicating which StrictDoc version the cache belongs to.
        # This helps prevent issues when switching between versions that may use
        # incompatible cache schemas.
        dir_for_sdoc_cache = os.path.join(dir_for_sdoc_cache, __version__)

        self.dir_for_sdoc_cache: str = dir_for_sdoc_cache

        self.project_features: List[str] = project_features
        self.server_host: str = server_host
        self.server_port: int = server_port
        self.include_doc_paths: List[str] = include_doc_paths
        self.exclude_doc_paths: List[str] = exclude_doc_paths
        self.source_root_path: Optional[str] = source_root_path
        self.include_source_paths: List[str] = include_source_paths
        self.exclude_source_paths: List[str] = exclude_source_paths
        self.test_report_root_dict: Dict[str, str] = test_report_root_dict
        self.source_nodes: List[Dict[str, str]] = source_nodes

        # Settings derived from the command-line parameters.

        # Common settings.
        self.input_paths: Optional[List[str]] = None
        self.output_dir: str = "output"

        # Export action.
        self.export_output_html_root: str = os.path.join(
            self.output_dir, "html"
        )
        self.export_formats: Optional[List[str]] = None
        self.export_included_documents: bool = False
        self.generate_bundle_document: bool = False
        self.filter_requirements: Optional[str] = None
        self.filter_sections: Optional[str] = None

        self.excel_export_fields: Optional[List[str]] = None

        self.html2pdf_template: Optional[str] = html2pdf_template
        self.bundle_document_version: Optional[str] = bundle_document_version
        self.bundle_document_date: Optional[str] = bundle_document_date

        self.traceability_matrix_relation_columns: Optional[
            List[Tuple[str, Optional[str]]]
        ] = traceability_matrix_relation_columns
        self.reqif_profile: str = reqif_profile
        self.reqif_multiline_is_xhtml: bool = reqif_multiline_is_xhtml
        self.reqif_enable_mid: bool = reqif_enable_mid
        self.reqif_import_markup: Optional[str] = reqif_import_markup

        #
        # auto_uid_mode: default is False. The True-case is used by the
        # manage/auto_uid command: the SDocNodeValidator will
        # not raise an exception if it sees a node with a missing UID.
        # Important for a special case:
        # The Manage UID command auto-generates the UID, so the field presence
        # validation has to be relaxed.
        # The GitHub issue report:
        # manage auto-uid: UID field REQUIRED True leads to an error
        # https://github.com/strictdoc-project/strictdoc/issues/1896
        #
        self.auto_uid_mode = False
        self.autouuid_include_sections: bool = False

        self.config_last_update: Optional[datetime.datetime] = (
            config_last_update
        )
        self.is_running_on_server: bool = False
        self.view: Optional[str] = None
        self.chromedriver: Optional[str] = chromedriver
        self.section_behavior: Optional[str] = section_behavior

        self.statistics_generator: Optional[str] = statistics_generator

    @staticmethod
    def default_config(environment: SDocRuntimeEnvironment) -> "ProjectConfig":
        assert isinstance(environment, SDocRuntimeEnvironment)
        return ProjectConfig(
            environment=environment,
            project_title=ProjectConfig.DEFAULT_PROJECT_TITLE,
            dir_for_sdoc_assets=ProjectConfig.DEFAULT_DIR_FOR_SDOC_ASSETS,
            dir_for_sdoc_cache=ProjectConfig.DEFAULT_DIR_FOR_SDOC_CACHE,
            project_features=ProjectConfig.DEFAULT_FEATURES,
            server_host=ProjectConfig.DEFAULT_SERVER_HOST,
            server_port=ProjectConfig.DEFAULT_SERVER_PORT,
            include_doc_paths=[],
            exclude_doc_paths=[],
            source_root_path=None,
            include_source_paths=[],
            exclude_source_paths=[],
            test_report_root_dict={},
            source_nodes=[],
            html2pdf_template=None,
            bundle_document_version=ProjectConfig.DEFAULT_BUNDLE_DOCUMENT_VERSION,
            bundle_document_date=ProjectConfig.DEFAULT_BUNDLE_DOCUMENT_COMMIT_DATE,
            traceability_matrix_relation_columns=None,
            reqif_profile=ReqIFProfile.P01_SDOC,
            reqif_multiline_is_xhtml=False,
            reqif_enable_mid=False,
            reqif_import_markup=None,
            config_last_update=None,
            chromedriver=None,
            section_behavior=ProjectConfig.DEFAULT_SECTION_BEHAVIOR,
            statistics_generator=None,
        )

    # Some server command settings can override the project config settings.
    def integrate_server_config(
        self, server_config: ServerCommandConfig
    ) -> None:
        self.is_running_on_server = True
        if (server_host_ := server_config.host) is not None:
            self.server_host = server_host_
        if (server_port_ := server_config.port) is not None:
            self.server_port = server_port_

        self.input_paths = [server_config.get_full_input_path()]
        if self.source_root_path is None:
            source_root_path = self.input_paths[0]
            # If the input argument is a relative path, convert it to an
            # absolute path.
            source_root_path = os.path.abspath(source_root_path)
            source_root_path = source_root_path.rstrip("/")
            self.source_root_path = source_root_path

        if (output_dir_ := server_config.output_path) is not None:
            self.output_dir = output_dir_
            self.export_output_html_root = os.path.join(output_dir_, "html")
            # If a custom cache folder is not specified in the config, adjust the
            # cache folder to be located in the output folder.
            if self.dir_for_sdoc_cache.startswith(
                ProjectConfig.DEFAULT_DIR_FOR_SDOC_CACHE
            ):
                self.dir_for_sdoc_cache = os.path.join(
                    output_dir_, "_cache", __version__
                )

        self.export_formats = ["html"]
        self.generate_bundle_document = False
        self.export_included_documents = True

    def integrate_export_config(
        self, export_config: ExportCommandConfig
    ) -> None:
        if export_config.project_title is not None:
            self.project_title = export_config.project_title

        self.input_paths = export_config.input_paths
        if self.source_root_path is None:
            source_root_path = export_config.input_paths[0]
            # If the input argument is a relative path, convert it to an
            # absolute path.
            source_root_path = os.path.abspath(source_root_path)
            source_root_path = source_root_path.rstrip("/")
            self.source_root_path = source_root_path

        #
        # Adjust the default output dir to the user-provided dir if needed.
        #
        output_dir = self.output_dir
        if export_config.output_dir is not None:
            output_dir = export_config.output_dir
        if not os.path.isabs(output_dir):
            cwd = os.getcwd()
            output_dir = os.path.join(cwd, output_dir)
        self.output_dir = output_dir

        # If a custom cache folder is not specified in the config, adjust the
        # cache folder to be located in the output folder.
        if self.dir_for_sdoc_cache.startswith(
            ProjectConfig.DEFAULT_DIR_FOR_SDOC_CACHE
        ):
            self.dir_for_sdoc_cache = os.path.join(
                output_dir, "_cache", __version__
            )

        self.export_output_html_root = os.path.join(self.output_dir, "html")
        self.export_formats = export_config.formats
        self.export_included_documents = export_config.included_documents
        self.generate_bundle_document = export_config.generate_bundle_document
        self.filter_requirements = export_config.filter_requirements
        self.filter_sections = export_config.filter_sections
        self.excel_export_fields = export_config.fields
        self.view = export_config.view
        self.chromedriver = export_config.chromedriver

        if (
            export_config.enable_mathjax
            and ProjectFeature.MATHJAX not in self.project_features
        ):
            self.project_features.append(ProjectFeature.MATHJAX)

        if export_config.reqif_profile is not None:
            self.reqif_profile = export_config.reqif_profile

        # If the TOML file sets this to True, ignore what is in CLI.
        if not self.reqif_multiline_is_xhtml:
            self.reqif_multiline_is_xhtml = (
                export_config.reqif_multiline_is_xhtml
            )
        if not self.reqif_enable_mid:
            self.reqif_enable_mid = export_config.reqif_enable_mid

    def is_feature_activated(self, feature: ProjectFeature) -> bool:
        return feature in self.project_features

    def is_activated_table_screen(self) -> bool:
        return ProjectFeature.TABLE_SCREEN in self.project_features

    def is_activated_trace_screen(self) -> bool:
        return ProjectFeature.TRACEABILITY_SCREEN in self.project_features

    def is_activated_deep_trace_screen(self) -> bool:
        return ProjectFeature.DEEP_TRACEABILITY_SCREEN in self.project_features

    def is_activated_project_statistics(self) -> bool:
        return ProjectFeature.PROJECT_STATISTICS_SCREEN in self.project_features

    def is_activated_requirements_to_source_traceability(self) -> bool:
        return (
            ProjectFeature.REQUIREMENT_TO_SOURCE_TRACEABILITY
            in self.project_features
        )

    def is_activated_requirements_coverage(self) -> bool:
        return (
            ProjectFeature.TRACEABILITY_MATRIX_SCREEN in self.project_features
        )

    def is_activated_standalone_document(self) -> bool:
        return (
            ProjectFeature.STANDALONE_DOCUMENT_SCREEN in self.project_features
        )

    def is_activated_search(self) -> bool:
        return (
            self.is_running_on_server
            and ProjectFeature.SEARCH in self.project_features
        )

    def is_activated_html2pdf(self) -> bool:
        return ProjectFeature.HTML2PDF in self.project_features

    def is_activated_diff(self) -> bool:
        return ProjectFeature.DIFF in self.project_features

    def is_activated_reqif(self) -> bool:
        return ProjectFeature.REQIF in self.project_features

    def is_activated_mathjax(self) -> bool:
        return ProjectFeature.MATHJAX in self.project_features

    def is_activated_mermaid(self) -> bool:
        return ProjectFeature.MERMAID in self.project_features

    def is_activated_rapidoc(self) -> bool:
        return ProjectFeature.RAPIDOC in self.project_features

    def is_activated_source_file_language_parsers(self) -> bool:
        return (
            ProjectFeature.SOURCE_FILE_LANGUAGE_PARSERS in self.project_features
        )

    def is_new_section_behavior(self) -> bool:
        return self.section_behavior == "[[SECTION]]"

    def get_strictdoc_root_path(self) -> str:
        return self.environment.path_to_strictdoc

    def get_path_to_cache_dir(self) -> str:
        return self.dir_for_sdoc_cache

    def get_static_files_path(self) -> str:
        return self.environment.get_static_files_path()

    def get_extra_static_files_path(self) -> str:
        return self.environment.get_extra_static_files_path()

    def get_project_hash(self) -> str:
        assert self.input_paths is not None and len(self.input_paths) > 0
        return get_md5(self.input_paths[0])

    def shall_parse_nodes(self, path_to_file: str) -> bool:
        if self.source_root_path is None:
            return False

        for sdoc_source_config_entry_ in self.source_nodes:
            # FIXME: Move the setting of full paths to .finalize() of this config
            #        class when it is implemented.
            full_path = sdoc_source_config_entry_.setdefault(
                "full_path",
                os.path.join(
                    self.source_root_path, sdoc_source_config_entry_["path"]
                ),
            )
            if path_to_file.startswith(full_path):
                return True

        return False


class ProjectConfigLoader:
    @staticmethod
    def load_from_path_or_get_default(
        *,
        path_to_config: str,
        environment: SDocRuntimeEnvironment,
    ) -> ProjectConfig:
        if not os.path.exists(path_to_config):
            return ProjectConfig.default_config(environment=environment)
        if os.path.isdir(path_to_config):
            path_to_config = os.path.join(path_to_config, "strictdoc.toml")
        if not os.path.isfile(path_to_config):
            return ProjectConfig.default_config(environment=environment)

        try:
            config_content = toml.load(path_to_config)
        except toml.decoder.TomlDecodeError as exception:
            raise StrictDocException(  # noqa: T201
                f"Could not parse the config file {path_to_config}: "
                f"{exception}."
            ) from None
        except Exception as exception:  # pragma: no cover
            raise AssertionError from exception

        config_last_update = get_file_modification_time(path_to_config)

        return ProjectConfigLoader._load_from_dictionary(
            config_dict=config_content,
            environment=environment,
            config_last_update=config_last_update,
            path_to_config=path_to_config,
        )

    @staticmethod
    def load_from_string(
        *, toml_string: str, environment: SDocRuntimeEnvironment
    ) -> ProjectConfig:
        config_dict = toml.loads(toml_string)
        return ProjectConfigLoader._load_from_dictionary(
            config_dict=config_dict,
            environment=environment,
            config_last_update=None,
            path_to_config=None,
        )

    @staticmethod
    def _load_from_dictionary(
        *,
        config_dict: Dict[str, Any],
        environment: SDocRuntimeEnvironment,
        config_last_update: Optional[datetime.datetime],
        path_to_config: Optional[str],
    ) -> ProjectConfig:
        if path_to_config is not None:
            assert os.path.isfile(path_to_config), path_to_config
            path_to_config = os.path.abspath(path_to_config)

        project_title = ProjectConfig.DEFAULT_PROJECT_TITLE
        dir_for_sdoc_assets = ProjectConfig.DEFAULT_DIR_FOR_SDOC_ASSETS
        dir_for_sdoc_cache = ProjectConfig.DEFAULT_DIR_FOR_SDOC_CACHE
        project_features = ProjectConfig.DEFAULT_FEATURES
        server_host = ProjectConfig.DEFAULT_SERVER_HOST
        server_port = ProjectConfig.DEFAULT_SERVER_PORT
        include_doc_paths: List[str] = []
        exclude_doc_paths: List[str] = []
        source_root_path = None
        include_source_paths: List[str] = []
        exclude_source_paths: List[str] = []
        test_report_root_dict: Dict[str, str] = {}
        source_nodes: List[Dict[str, str]] = []
        html2pdf_template: Optional[str] = None
        bundle_document_version = ProjectConfig.DEFAULT_BUNDLE_DOCUMENT_VERSION
        bundle_document_date = ProjectConfig.DEFAULT_BUNDLE_DOCUMENT_COMMIT_DATE

        traceability_matrix_relation_columns: Optional[
            List[Tuple[str, Optional[str]]]
        ] = None
        reqif_profile = ReqIFProfile.P01_SDOC
        reqif_multiline_is_xhtml = False
        reqif_enable_mid = False
        reqif_import_markup: Optional[str] = None
        chromedriver: Optional[str] = None

        section_behavior: str = ProjectConfig.DEFAULT_SECTION_BEHAVIOR
        statistics_generator: Optional[str] = None

        if "project" in config_dict:
            project_content = config_dict["project"]
            project_title = project_content.get("title", project_title)
            dir_for_sdoc_assets = project_content.get(
                "html_assets_strictdoc_dir", dir_for_sdoc_assets
            )
            dir_for_sdoc_cache = project_content.get(
                "cache_dir", dir_for_sdoc_cache
            )

            project_features = project_content.get("features", project_features)
            if not isinstance(project_features, list):
                print(  # noqa: T201
                    f"error: strictdoc.toml: 'feature' parameter must be an "
                    f"array: '{project_features}'."
                )
                sys.exit(1)

            for feature in project_features:
                if feature not in ProjectFeature.all():
                    print(  # noqa: T201
                        f"error: strictdoc.toml: unknown feature declared: "
                        f"'{feature}'."
                    )
                    sys.exit(1)
            if ProjectFeature.ALL_FEATURES in project_features:
                project_features = ProjectFeature.all()

            statistics_generator = project_content.get(
                "statistics_generator", statistics_generator
            )

            include_doc_paths = project_content.get(
                "include_doc_paths", include_doc_paths
            )
            assert isinstance(include_doc_paths, list)
            for include_doc_path in include_doc_paths:
                try:
                    validate_mask(include_doc_path)
                except SyntaxError as exception_:
                    raise SyntaxError(
                        f"strictdoc.toml: 'include_doc_paths': "
                        f"{exception_} Provided string: '{include_doc_path}'."
                    ) from exception_

            exclude_doc_paths = project_content.get(
                "exclude_doc_paths", exclude_doc_paths
            )
            assert isinstance(exclude_doc_paths, list)
            for exclude_doc_path in exclude_doc_paths:
                try:
                    validate_mask(exclude_doc_path)
                except SyntaxError as exception_:
                    raise SyntaxError(
                        f"strictdoc.toml: 'exclude_doc_paths': "
                        f"{exception_} Provided string: '{exclude_doc_path}'."
                    ) from exception_

            source_root_path = project_content.get(
                "source_root_path", source_root_path
            )
            if source_root_path is not None:
                original_source_root_path = source_root_path
                if (
                    not os.path.isabs(source_root_path)
                    and path_to_config is not None
                ):
                    source_root_path = os.path.join(
                        os.path.dirname(path_to_config), source_root_path
                    )
                    source_root_path = os.path.abspath(source_root_path)
                if not os.path.isdir(source_root_path):
                    raise ValueError(
                        f"strictdoc.toml: 'source_root_path': "
                        f"Provided path does not exist: "
                        f"{original_source_root_path}."
                    )
                if not os.path.isabs(source_root_path):
                    source_root_path = os.path.abspath(source_root_path)
            include_source_paths = project_content.get(
                "include_source_paths", include_source_paths
            )
            assert isinstance(include_source_paths, list)
            for include_source_path in include_source_paths:
                try:
                    validate_mask(include_source_path)
                except SyntaxError as exception_:
                    raise SyntaxError(
                        f"strictdoc.toml: 'include_source_paths': "
                        f"{exception_} Provided string: '{include_source_path}'."
                    ) from exception_

            exclude_source_paths = project_content.get(
                "exclude_source_paths", exclude_source_paths
            )
            assert isinstance(exclude_source_paths, list)
            for exclude_source_path in exclude_source_paths:
                try:
                    validate_mask(exclude_source_path)
                except SyntaxError as exception_:
                    raise SyntaxError(
                        f"strictdoc.toml: 'exclude_source_paths': "
                        f"{exception_} Provided string: '{exclude_source_path}'."
                    ) from exception_

            html2pdf_template = project_content.get(
                "html2pdf_template", html2pdf_template
            )
            if html2pdf_template is not None:
                assert not os.path.isabs(html2pdf_template)
                if path_to_config is not None:
                    path_to_config_dir = os.path.dirname(path_to_config)
                    html2pdf_template = os.path.join(
                        path_to_config_dir, html2pdf_template
                    )
                if not os.path.isfile(html2pdf_template):
                    raise ValueError(
                        "strictdoc.toml: 'html2pdf_template': "
                        f"invalid path to a template file: {html2pdf_template}."
                    )

            bundle_document_version = project_content.get(
                "bundle_document_version", bundle_document_version
            )

            bundle_document_date = project_content.get(
                "bundle_document_date", bundle_document_date
            )

            traceability_matrix_relation_columns_config: Optional[List[str]] = (
                project_content.get(
                    "traceability_matrix_relation_columns", None
                )
            )
            if traceability_matrix_relation_columns_config is not None:
                assert isinstance(
                    traceability_matrix_relation_columns_config, list
                )
                traceability_matrix_relation_columns = []
                for (
                    relation_column_string_
                ) in traceability_matrix_relation_columns_config:
                    relation_tuple = parse_relation_tuple(
                        relation_column_string_
                    )
                    assert relation_tuple is not None
                    traceability_matrix_relation_columns.append(relation_tuple)

            chromedriver = project_content.get("chromedriver", chromedriver)
            if chromedriver is not None and not os.path.isfile(chromedriver):
                raise ValueError(
                    f"strictdoc.toml: 'chromedriver': "
                    f"not found at path: {chromedriver}."
                )
            if (
                test_report_root_dict_ := project_content.get(
                    "test_report_root_dict", None
                )
            ) is not None:
                assert isinstance(test_report_root_dict_, list), (
                    test_report_root_dict
                )
                for test_report_root_entry_ in test_report_root_dict_:
                    assert isinstance(test_report_root_entry_, dict)
                    test_report_root_dict.update(test_report_root_entry_)

            section_behavior = project_content.get(
                "section_behavior", section_behavior
            )
            assert section_behavior in ("[SECTION]", "[[SECTION]]")

            if "source_nodes" in project_content:
                source_nodes_config = project_content["source_nodes"]
                assert isinstance(source_nodes_config, list)
                for item_ in source_nodes_config:
                    source_node_path = next(iter(item_))
                    source_node_uid = item_[source_node_path]["uid"]
                    source_node_node_type = item_[source_node_path]["node_type"]
                    source_nodes.append(
                        {
                            "path": source_node_path,
                            "uid": source_node_uid,
                            "node_type": source_node_node_type,
                        }
                    )

        if "server" in config_dict:
            server_content = config_dict["server"]
            server_host = server_content.get("host", server_host)
            if not is_valid_host(server_host):
                raise ValueError(
                    f"strictdoc.toml: 'host': invalid host: {server_host}'."
                )

            server_port = server_content.get("port", server_port)
            if not (
                isinstance(server_port, int) and 1024 < server_port < 65000
            ):
                raise ValueError(
                    f"strictdoc.toml: 'port': invalid port: {server_port}'."
                )

        if "reqif" in config_dict:
            # FIXME: Introduce at least a basic validation.
            reqif_content = config_dict["reqif"]
            reqif_multiline_is_xhtml = reqif_content.get(
                "multiline_is_xhtml", False
            )
            assert isinstance(reqif_multiline_is_xhtml, bool), reqif_content

            # FIXME
            reqif_enable_mid = reqif_content.get("enable_mid", False)
            assert isinstance(reqif_enable_mid, bool), reqif_content

            # FIXME
            reqif_import_markup = reqif_content.get("import_markup", None)
            if reqif_import_markup is not None:
                assert reqif_import_markup in SDocMarkup.ALL, (
                    reqif_import_markup
                )

        return ProjectConfig(
            environment=environment,
            project_title=project_title,
            dir_for_sdoc_assets=dir_for_sdoc_assets,
            dir_for_sdoc_cache=dir_for_sdoc_cache,
            project_features=project_features,
            server_host=server_host,
            server_port=server_port,
            include_doc_paths=include_doc_paths,
            exclude_doc_paths=exclude_doc_paths,
            source_root_path=source_root_path,
            include_source_paths=include_source_paths,
            exclude_source_paths=exclude_source_paths,
            test_report_root_dict=test_report_root_dict,
            source_nodes=source_nodes,
            html2pdf_template=html2pdf_template,
            bundle_document_version=bundle_document_version,
            bundle_document_date=bundle_document_date,
            traceability_matrix_relation_columns=traceability_matrix_relation_columns,
            reqif_profile=reqif_profile,
            reqif_multiline_is_xhtml=reqif_multiline_is_xhtml,
            reqif_enable_mid=reqif_enable_mid,
            reqif_import_markup=reqif_import_markup,
            config_last_update=config_last_update,
            chromedriver=chromedriver,
            section_behavior=section_behavior,
            statistics_generator=statistics_generator,
        )



from typing import Dict, List


class ErrorObject:
    def __init__(self) -> None:
        self.errors: Dict[str, List[str]] = {}

    def any_errors(self) -> bool:
        return len(self.errors) > 0

    def get_errors(self, field_name: str) -> List[str]:
        if field_name not in self.errors:
            return []
        return self.errors[field_name]

    def add_error(self, field_name: str, error: str) -> None:
        error_container = self.errors.setdefault(field_name, [])
        error_container.append(error)



"""
@relation(SDOC-SRS-55, scope=file)
"""

from collections import defaultdict
from enum import Enum
from typing import Dict, Iterator, List, Optional, Set, Tuple, Union

from starlette.datastructures import FormData

from strictdoc.backend.sdoc.errors.document_tree_error import DocumentTreeError
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementField,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldSingleChoice,
    RequirementFieldType,
)
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    FileEntry,
    FileEntryFormat,
    FileReference,
    ParentReqReference,
    Reference,
)
from strictdoc.core.constants import GraphLinkType
from strictdoc.core.graph.abstract_bucket import ALL_EDGES
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.tree_cycle_detector import SingleShotTreeCycleDetector
from strictdoc.export.rst.rst_to_html_fragment_writer import (
    RstToHtmlFragmentWriter,
)
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.form_data import ParsedFormData, parse_form_data
from strictdoc.helpers.mid import MID
from strictdoc.helpers.string import sanitize_html_form_field
from strictdoc.server.error_object import ErrorObject


class RequirementFormFieldType(str, Enum):
    SINGLELINE = "SINGLELINE"
    MULTILINE = "MULTILINE"


@auto_described
class RequirementFormField:
    def __init__(
        self,
        field_mid: str,
        field_name: str,
        field_type: RequirementFormFieldType,
        field_value: str,
        field_gef_type: str = RequirementFieldType.STRING,
    ) -> None:
        assert isinstance(field_value, str)
        self.field_mid: str = field_mid
        self.field_name: str = field_name
        self.field_value: str = field_value
        self.field_type = field_type
        self.field_gef_type: str = field_gef_type

    def is_multiline(self) -> bool:
        return self.field_type == RequirementFormFieldType.MULTILINE

    def is_autocompletable(self) -> bool:
        return self.field_gef_type in (
            RequirementFieldType.SINGLE_CHOICE,
            RequirementFieldType.MULTIPLE_CHOICE,
            RequirementFieldType.TAG,
        )

    def is_multiplechoice(self) -> bool:
        return self.field_gef_type in (
            RequirementFieldType.MULTIPLE_CHOICE,
            RequirementFieldType.TAG,
        )

    def get_input_field_name(self) -> str:
        return f"requirement[fields][{self.field_mid}][value]"

    def get_input_field_type_name(self) -> str:
        return f"requirement[fields][{self.field_mid}][name]"

    @staticmethod
    def create_from_grammar_field(
        *,
        grammar_field: GrammarElementField,
        multiline: bool,
        value: str,
    ) -> "RequirementFormField":
        assert isinstance(value, str), (
            grammar_field,
            multiline,
            value,
        )
        if grammar_field.gef_type in (
            RequirementFieldType.STRING,
            RequirementFieldType.SINGLE_CHOICE,
            RequirementFieldType.MULTIPLE_CHOICE,
            RequirementFieldType.TAG,
        ):
            return RequirementFormField(
                field_mid=MID.create(),
                field_name=grammar_field.title,
                field_type=(
                    RequirementFormFieldType.MULTILINE
                    if multiline
                    else RequirementFormFieldType.SINGLELINE
                ),
                field_value=value,
                field_gef_type=grammar_field.gef_type,
            )
        raise NotImplementedError(grammar_field)

    @staticmethod
    def create_existing_from_grammar_field(
        grammar_field: GrammarElementField,
        multiline: bool,
        requirement_field: SDocNodeField,
    ) -> "RequirementFormField":
        if grammar_field.gef_type in (
            RequirementFieldType.STRING,
            RequirementFieldType.SINGLE_CHOICE,
            RequirementFieldType.MULTIPLE_CHOICE,
            RequirementFieldType.TAG,
        ):
            field_value = requirement_field.get_text_value()
            return RequirementFormField(
                field_mid=MID.create(),
                field_name=grammar_field.title,
                field_type=(
                    RequirementFormFieldType.MULTILINE
                    if multiline
                    else RequirementFormFieldType.SINGLELINE
                ),
                field_value=field_value,
                field_gef_type=grammar_field.gef_type,
            )
        raise NotImplementedError(grammar_field)


@auto_described
class RequirementReferenceFormField:
    class FieldType(str, Enum):
        PARENT = "Parent"
        CHILD = "Child"
        FILE = "File"

    def __init__(
        self,
        field_mid: str,
        field_type: FieldType,
        field_value: str,
        field_role: Optional[str],
    ) -> None:
        assert isinstance(field_mid, str), field_mid
        assert isinstance(field_value, str), field_value
        self.field_mid: str = field_mid
        self.field_type = field_type
        self.field_value: str = field_value
        self.field_role: str = (
            field_role if field_role is not None and len(field_role) > 0 else ""
        )
        self.validation_messages: List[str] = []

    def get_input_field_name(self) -> str:
        return "requirement_relation"

    def get_value_field_name(self) -> str:
        return f"requirement[relations][{self.field_mid}][value]"

    def get_type_field_name(self) -> str:
        return f"requirement[relations][{self.field_mid}][typerole]"


@auto_described
class RequirementFormObject(ErrorObject):
    """
    Class for managing node fields in the StrictDoc web editor.

    context_document_mid: The MID of the document where the requirement is edited.
                          Normally, this is the requirement's own document but can
                          also be the parent document if requirement's own document
                          is included to it.
    """

    def __init__(
        self,
        *,
        is_new: bool,
        element_type: str,
        requirement_mid: str,
        document_mid: str,
        context_document_mid: str,
        fields: List[RequirementFormField],
        reference_fields: List[RequirementReferenceFormField],
        existing_requirement_uid: Optional[str],
        grammar: DocumentGrammar,
        # FIXME: Better name.
        relation_types: List[str],
    ) -> None:
        super().__init__()
        assert isinstance(element_type, str), element_type
        self.is_new: bool = is_new
        self.element_type: str = element_type
        self.requirement_mid: str = requirement_mid
        self.document_mid: str = document_mid
        self.context_document_mid: str = context_document_mid
        fields_dict: Dict[str, List[RequirementFormField]] = {}
        for field in fields:
            fields_dict.setdefault(field.field_name, []).append(field)

        self.fields: Dict[str, List[RequirementFormField]] = fields_dict
        self.reference_fields: List[RequirementReferenceFormField] = (
            reference_fields
        )
        self.existing_requirement_uid: Optional[str] = existing_requirement_uid
        self.grammar: DocumentGrammar = grammar
        self.relation_types: List[str] = relation_types

    @staticmethod
    def create_from_request(
        *,
        is_new: bool,
        requirement_mid: str,
        request_form_data: FormData,
        document: SDocDocument,
        existing_requirement_uid: Optional[str],
    ) -> "RequirementFormObject":
        request_form_data_as_list = [
            (field_name, field_value)
            for field_name, field_value in request_form_data.multi_items()
        ]
        request_form_dict: ParsedFormData = assert_cast(
            parse_form_data(request_form_data_as_list), dict
        )
        requirement_fields = defaultdict(list)
        form_ref_fields: List[RequirementReferenceFormField] = []

        context_document_mid = assert_cast(
            request_form_dict["context_document_mid"], str
        )
        requirement_dict = assert_cast(request_form_dict["requirement"], dict)

        element_type = assert_cast(request_form_dict["element_type"], str)
        requirement_fields_dict = assert_cast(requirement_dict["fields"], dict)
        for _, field_dict in requirement_fields_dict.items():
            assert isinstance(field_dict, dict), type(field_dict)

            field_name = field_dict["name"]
            field_value = field_dict["value"]
            requirement_fields[field_name].append(field_value)

        # FIXME: defaulting to {}
        requirement_relations_dict = requirement_dict.get("relations", {})
        for relation_mid, relation_dict in requirement_relations_dict.items():
            # FIXME: Editing files is not supported. Fix this hack ASAP.
            relation_typerole = relation_dict.get("typerole", "File")
            relation_typerole_parts = relation_typerole.split(",")
            if len(relation_typerole_parts) == 2:
                relation_type = relation_typerole_parts[0]
                relation_role = relation_typerole_parts[1]
            elif len(relation_typerole_parts) == 1:
                relation_type = relation_typerole_parts[0]
                relation_role = None
            else:
                raise AssertionError("Must not reach here")  # pragma: no cover

            field_type = {
                "Parent": RequirementReferenceFormField.FieldType.PARENT,
                "Child": RequirementReferenceFormField.FieldType.CHILD,
                "File": RequirementReferenceFormField.FieldType.FILE,
            }[relation_type]

            relation_value = relation_dict["value"].strip()

            form_ref_fields.append(
                RequirementReferenceFormField(
                    field_mid=relation_mid,
                    field_type=field_type,
                    field_value=relation_value,
                    field_role=relation_role,
                )
            )

        assert document.grammar is not None
        grammar: DocumentGrammar = document.grammar
        element: GrammarElement = grammar.elements_by_type[element_type]
        form_fields: List[RequirementFormField] = []

        for field_idx, field_name in enumerate(element.field_titles):
            multiline = element.is_field_idx_multiline(field_idx)

            field = element.fields_map[field_name]

            if field_name not in requirement_fields:
                continue

            requirement_field_values = requirement_fields.get(field_name, [])
            for requirement_field_value in requirement_field_values:
                sanitized_field_value: str = sanitize_html_form_field(
                    requirement_field_value, multiline=multiline
                )
                form_field = RequirementFormField.create_from_grammar_field(
                    grammar_field=field,
                    multiline=multiline,
                    value=sanitized_field_value,
                )
                form_fields.append(form_field)

        form_object = RequirementFormObject(
            is_new=is_new,
            element_type=element_type,
            requirement_mid=requirement_mid,
            document_mid=document.reserved_mid,
            context_document_mid=context_document_mid,
            fields=form_fields,
            reference_fields=form_ref_fields,
            existing_requirement_uid=existing_requirement_uid,
            grammar=grammar,
            relation_types=element.get_relation_types(),
        )
        return form_object

    @staticmethod
    def create_new(
        *,
        document: SDocDocument,
        context_document_mid: str,
        next_uid: Optional[str],
        element_type: str,
    ) -> "RequirementFormObject":
        """
        Create a new node form object.

        For now, the next_uid cannot be non-None for TEXT nodes. This will likely
        change in the future.
        """

        assert document.grammar is not None

        new_requirement_mid: MID = MID.create()

        grammar: DocumentGrammar = document.grammar
        element: GrammarElement = grammar.elements_by_type[element_type]

        form_fields: List[RequirementFormField] = []

        for field_idx, field_name in enumerate(element.field_titles):
            field = element.fields_map[field_name]
            multiline = element.is_field_idx_multiline(field_idx)
            form_field: RequirementFormField = (
                RequirementFormField.create_from_grammar_field(
                    grammar_field=field,
                    multiline=multiline,
                    value="",
                )
            )
            form_fields.append(form_field)
            if form_field.field_name == "UID" and next_uid is not None:
                form_field.field_value = next_uid
            elif form_field.field_name == "MID" and document.config.enable_mid:
                form_field.field_value = new_requirement_mid.get_string_value()

        return RequirementFormObject(
            is_new=True,
            element_type=element_type,
            requirement_mid=new_requirement_mid,
            document_mid=document.reserved_mid,
            context_document_mid=context_document_mid,
            fields=form_fields,
            reference_fields=[],
            existing_requirement_uid=None,
            grammar=grammar,
            relation_types=element.get_relation_types(),
        )

    @staticmethod
    def create_from_requirement(
        *,
        requirement: SDocNode,
        context_document_mid: str,
    ) -> "RequirementFormObject":
        assert isinstance(requirement, SDocNode)
        document: SDocDocument = assert_cast(
            requirement.get_document(), SDocDocument
        )
        assert document.grammar is not None
        grammar: DocumentGrammar = document.grammar
        element: GrammarElement = grammar.elements_by_type[
            requirement.node_type
        ]

        grammar_element_relations = element.get_relation_types()

        form_fields: List[RequirementFormField] = []
        form_refs_fields: List[RequirementReferenceFormField] = []

        for field_idx, field_name in enumerate(element.field_titles):
            multiline = element.is_field_idx_multiline(field_idx)

            # Handle all other fields in a general way.
            field = element.fields_map[field_name]

            if field_name in requirement.ordered_fields_lookup:
                for requirement_field in requirement.ordered_fields_lookup[
                    field_name
                ]:
                    form_field = (
                        RequirementFormField.create_existing_from_grammar_field(
                            field,
                            multiline=multiline,
                            requirement_field=requirement_field,
                        )
                    )
                    form_fields.append(form_field)
            else:
                form_field = RequirementFormField.create_from_grammar_field(
                    grammar_field=field,
                    multiline=multiline,
                    value="",
                )
                form_fields.append(form_field)

        for reference_value in requirement.relations:
            if isinstance(reference_value, ParentReqReference):
                parent_reference: ParentReqReference = reference_value
                form_ref_field = RequirementReferenceFormField(
                    field_mid=parent_reference.mid,
                    field_type=RequirementReferenceFormField.FieldType.PARENT,
                    field_value=parent_reference.ref_uid,
                    field_role=parent_reference.role,
                )
                form_refs_fields.append(form_ref_field)
            elif isinstance(reference_value, ChildReqReference):
                child_req_reference: ChildReqReference = reference_value
                form_ref_field = RequirementReferenceFormField(
                    field_mid=child_req_reference.mid,
                    field_type=RequirementReferenceFormField.FieldType.CHILD,
                    field_value=child_req_reference.ref_uid,
                    field_role=child_req_reference.role,
                )
                form_refs_fields.append(form_ref_field)
            elif isinstance(reference_value, FileReference):
                child_file_reference: FileReference = reference_value
                form_ref_field = RequirementReferenceFormField(
                    field_mid=child_file_reference.mid,
                    field_type=RequirementReferenceFormField.FieldType.FILE,
                    field_value=child_file_reference.get_posix_path(),
                    field_role=child_file_reference.role,
                )
                form_refs_fields.append(form_ref_field)
        return RequirementFormObject(
            is_new=False,
            element_type=requirement.node_type,
            requirement_mid=requirement.reserved_mid,
            document_mid=document.reserved_mid,
            context_document_mid=context_document_mid,
            fields=form_fields,
            reference_fields=form_refs_fields,
            existing_requirement_uid=requirement.reserved_uid,
            grammar=grammar,
            relation_types=grammar_element_relations,
        )

    @staticmethod
    def clone_from_requirement(
        *, requirement: SDocNode, context_document_mid: str, clone_uid: str
    ) -> "RequirementFormObject":
        form_object: RequirementFormObject = (
            RequirementFormObject.create_from_requirement(
                requirement=requirement,
                context_document_mid=context_document_mid,
            )
        )
        for field_name, fields_ in form_object.fields.items():
            if field_name == "UID":
                field: RequirementFormField = fields_[0]
                field.field_value = clone_uid
        form_object.requirement_mid = MID.create()

        return form_object

    def any_errors(self) -> bool:
        if super().any_errors():
            return True
        for reference_field in self.reference_fields:
            if len(reference_field.validation_messages) > 0:
                return True
        return False

    def get_requirement_relations(
        self, requirement: SDocNode
    ) -> List[Reference]:
        references: List[Reference] = []
        reference_field: RequirementReferenceFormField
        for reference_field in self.reference_fields:
            ref_uid = reference_field.field_value
            ref_type = reference_field.field_type
            ref_role = reference_field.field_role
            if ref_type == RequirementReferenceFormField.FieldType.PARENT:
                references.append(
                    ParentReqReference(
                        parent=requirement, ref_uid=ref_uid, role=ref_role
                    )
                )
            elif ref_type == RequirementReferenceFormField.FieldType.CHILD:
                references.append(
                    ChildReqReference(
                        parent=requirement, ref_uid=ref_uid, role=ref_role
                    )
                )
            elif ref_type == RequirementReferenceFormField.FieldType.FILE:
                file_entry = FileEntry(
                    parent=requirement,
                    g_file_format=FileEntryFormat.SOURCECODE,
                    g_file_path=reference_field.field_value,
                    g_line_range="",
                    function=None,
                    clazz=None,
                )
                references.append(
                    FileReference(
                        parent=requirement,
                        g_file_entry=file_entry,
                    )
                )
            else:
                raise NotImplementedError(ref_type)
        return references

    def enumerate_fields(
        self, multiline: bool
    ) -> Iterator[List[RequirementFormField]]:
        for field_name_, field in self.fields.items():
            try:
                requirement_field: RequirementFormField = field[0]
            except IndexError as index_error_:
                raise AssertionError(
                    f"Expected field name to correspond to an existing field: {field_name_}."
                ) from index_error_

            if multiline:
                if not requirement_field.is_multiline():
                    continue
            else:
                if requirement_field.is_multiline():
                    continue
            yield field

    def enumerate_reference_fields(
        self,
    ) -> Iterator[RequirementReferenceFormField]:
        yield from self.reference_fields

    def enumerate_relation_roles(
        self, relation_field: RequirementReferenceFormField
    ) -> Iterator[Tuple[str, Optional[str], bool]]:
        requirement_element = self.grammar.elements_by_type[self.element_type]
        for relation_ in requirement_element.relations:
            is_current = (
                relation_field.field_type == relation_.relation_type
                and (
                    relation_field.field_role == relation_.relation_role
                    or (
                        relation_field.field_role == ""
                        and relation_.relation_role is None
                    )
                )
            )
            yield relation_.relation_type, relation_.relation_role, is_current

    def validate(
        self,
        *,
        traceability_index: TraceabilityIndex,
        context_document: SDocDocument,
        config: ProjectConfig,
    ) -> None:
        assert isinstance(traceability_index, TraceabilityIndex)
        assert isinstance(context_document, SDocDocument)

        #
        # Ensure that at least one field must be non-empty.
        #
        at_least_one_non_empty_field_present = False
        for field_fields_ in self.fields.values():
            for field_ in field_fields_:
                if len(field_.field_value) > 0:
                    at_least_one_non_empty_field_present = True
                    break
            if at_least_one_non_empty_field_present:
                break
        if not at_least_one_non_empty_field_present:
            self.add_error(
                "_GENERAL_",
                "At least one node field must be non-empty.",
            )

        #
        # MID uniqueness check.
        # FIXME: MID uniqueness if a node is updated.
        # """
        if self.is_new and "MID" in self.fields:
            new_node_mid = self.fields["MID"][0].field_value
            if len(new_node_mid) > 0:
                existing_node_with_this_mid = (
                    traceability_index.get_node_by_mid_weak(MID(new_node_mid))
                )
                if existing_node_with_this_mid is not None:
                    self.add_error(
                        "MID",
                        (
                            f"A node with this MID already exists, "
                            "please select another MID: "
                            f"{new_node_mid}."
                        ),
                    )

        #
        # UID uniqueness check.
        #
        new_node_uid_or_none: Optional[str] = None
        if "UID" in self.fields:
            new_node_uid = self.fields["UID"][0].field_value
            if len(new_node_uid) > 0:
                new_node_uid_or_none = new_node_uid

        if new_node_uid_or_none is not None and (
            self.is_new or self.existing_requirement_uid != new_node_uid_or_none
        ):
            existing_node_with_this_uid = (
                traceability_index.get_node_by_uid_weak(new_node_uid_or_none)
            )
            if existing_node_with_this_uid is not None:
                self.add_error(
                    "UID",
                    (
                        "The chosen UID must be unique. "
                        "Another node with this UID already exists: "
                        f"'{new_node_uid_or_none}'."
                    ),
                )

        #
        # Ensure that UID doesn't have any incoming links if it is going to be
        # renamed or removed.
        #
        if self.existing_requirement_uid is not None:
            if (
                new_node_uid_or_none is None
                or self.existing_requirement_uid != new_node_uid_or_none
            ):
                existing_node: SDocNode = traceability_index.get_node_by_mid(
                    MID(self.requirement_mid)
                )

                existing_incoming_links: Optional[List[InlineLink]] = (
                    traceability_index.get_incoming_links(existing_node)
                )
                if (
                    existing_incoming_links is not None
                    and len(existing_incoming_links) > 0
                ):
                    self.add_error(
                        "UID",
                        (
                            "Renaming a node UID when the node has "
                            "incoming links is not supported yet. "
                            "Please delete all incoming links first."
                        ),
                    )

        requirement_element = self.grammar.elements_by_type[self.element_type]

        for grammar_element_field_ in requirement_element.fields:
            if grammar_element_field_.gef_type == RequirementFieldType.STRING:
                if grammar_element_field_.title not in self.fields:
                    continue

                for form_field_ in self.fields[grammar_element_field_.title]:
                    field_value = form_field_.field_value

                    # If field not empty, validate its RST syntax.
                    if len(field_value) > 0:
                        (
                            parsed_html,
                            rst_error,
                        ) = RstToHtmlFragmentWriter(
                            project_config=config,
                            context_document=context_document,
                        ).write_with_validation(field_value)
                        if parsed_html is None:
                            assert rst_error is not None
                            self.add_error(
                                grammar_element_field_.title, rst_error
                            )
                    # If field is empty, check if required and validate for emptiness.
                    else:
                        if grammar_element_field_.required:
                            self.add_error(
                                grammar_element_field_.title,
                                (
                                    f"Node's {grammar_element_field_.title} must not be empty. "
                                    f"If there is no appropriate value for this field yet, "
                                    f"enter TBD (to be done) or TBC (to be confirmed)."
                                ),
                            )
                        continue

            elif grammar_element_field_.gef_type in (
                RequirementFieldType.SINGLE_CHOICE,
                RequirementFieldType.MULTIPLE_CHOICE,
            ):
                self._validate_choice(grammar_element_field_)

        requirement_uid: Optional[str] = (
            self.fields["UID"][0].field_value if "UID" in self.fields else None
        )
        if len(self.reference_fields) > 0 and (
            requirement_uid is None or len(requirement_uid) == 0
        ):
            self.add_error(
                "UID",
                "Requirement with parent relations must have an UID. "
                "Either provide a parent UID, or "
                "delete the parent requirement relations.",
            )

        if (
            self.existing_requirement_uid is not None
            and self.existing_requirement_uid != requirement_uid
        ):
            if len(self.reference_fields) > 0:
                self.add_error(
                    "UID",
                    "Not supported yet: "
                    "Renaming a requirement UID when the requirement has "
                    "parent requirement relations. For now, manually delete the "
                    "relations, rename the UID, recreate the relations.",
                )

            existing_node = assert_cast(
                traceability_index.get_node_by_uid_weak(
                    self.existing_requirement_uid
                ),
                SDocNode,
            )

            if traceability_index.has_children_requirements(existing_node):
                self.add_error(
                    "UID",
                    "Not supported yet: "
                    "Renaming a requirement UID when the requirement has "
                    "child requirement relations. For now, manually delete the "
                    "relations, rename the UID, recreate the relations.",
                )
                return

        if requirement_uid is not None:
            relation_target_uids_so_far: Set[str] = set()
            for reference_field in self.reference_fields:
                if reference_field.field_type in ("Parent", "Child"):
                    link_uid = reference_field.field_value
                    if len(link_uid) == 0:
                        reference_field.validation_messages.append(
                            "Requirement relation UID must not be empty."
                        )
                        continue
                    elif not traceability_index.has_node_connections(link_uid):
                        reference_field.validation_messages.append(
                            f'Parent requirement with an UID "{link_uid}" '
                            f"does not exist."
                        )
                        continue

                    # Validate that every UID can be only referenced once.
                    if link_uid in relation_target_uids_so_far:
                        reference_field.validation_messages.append(
                            f'A target requirement with a UID "{link_uid}" '
                            "is referenced more than once. Multiple relations "
                            "to the same target requirement are not allowed."
                        )
                        continue
                    relation_target_uids_so_far.add(link_uid)

                    # Check if the target document supports a given relation.
                    node_grammar_element: GrammarElement = (
                        self.grammar.elements_by_type[self.element_type]
                    )
                    field_role_or_none = (
                        reference_field.field_role
                        if reference_field.field_role is not None
                        and len(reference_field.field_role) > 0
                        else None
                    )

                    # This is not a realistic case to happen when a node is
                    # edited in UI because the UI dropdown element whitelists
                    # the available relation types.
                    # Using an assert anyway just to make sure.
                    assert node_grammar_element.has_relation_type_role(
                        reference_field.field_type, field_role_or_none
                    )

                    self._validate_no_cycle_by_new_node(
                        traceability_index, reference_field, requirement_uid
                    )

    @staticmethod
    def _validate_no_cycle_by_new_node(
        traceability_index: TraceabilityIndex,
        reference_field: RequirementReferenceFormField,
        requirement_uid: str,
    ) -> None:
        """
        Check if a relation being added by a new node would form a cycle.
        """

        ref_uid = reference_field.field_value

        def parent_lambda(node_id: str) -> List[str]:
            node = traceability_index.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=node_id,
            )
            return list(
                map(
                    lambda node_: node_.reserved_uid,
                    traceability_index.graph_database.get_link_values(
                        link_type=GraphLinkType.NODE_TO_PARENT_NODES,
                        lhs_node=node,
                        edge=ALL_EDGES,
                    ),
                )
            )

        def child_lambda(node_id: str) -> List[str]:
            node = traceability_index.graph_database.get_link_value(
                link_type=GraphLinkType.UID_TO_NODE,
                lhs_node=node_id,
            )
            return list(
                map(
                    lambda node_: node_.reserved_uid,
                    traceability_index.graph_database.get_link_values(
                        link_type=GraphLinkType.NODE_TO_CHILD_NODES,
                        lhs_node=node,
                        edge=ALL_EDGES,
                    ),
                )
            )

        relations_lambda = (
            parent_lambda
            if reference_field.field_type == "Parent"
            else child_lambda
        )

        cycle_detector = SingleShotTreeCycleDetector()
        try:
            cycle_detector.check_node(
                requirement_uid,
                ref_uid,
                relations_lambda,
            )
        except DocumentTreeError as error_:
            reference_field.validation_messages.append(
                error_.to_validation_message()
            )

    def _validate_choice(
        self, grammar_element_field: GrammarElementField
    ) -> None:
        field_0 = self.fields[grammar_element_field.title][0]
        if len(field_0.field_value) == 0 and not grammar_element_field.required:
            # The empty choice fields are allowed if the field is not REQUIRED.
            return

        choice_grammar_element_field: Union[
            GrammarElementFieldSingleChoice,
            GrammarElementFieldMultipleChoice,
        ] = assert_cast(
            grammar_element_field,
            (
                GrammarElementFieldSingleChoice,
                GrammarElementFieldMultipleChoice,
            ),
        )
        if (
            grammar_element_field.gef_type == RequirementFieldType.SINGLE_CHOICE
            and field_0.field_value not in choice_grammar_element_field.options
        ):
            self.add_error(
                grammar_element_field.title,
                (
                    f"Node's {grammar_element_field.title} must be a value one of "
                    f"{', '.join(choice_grammar_element_field.options)}."
                ),
            )
        elif (
            grammar_element_field.gef_type
            == RequirementFieldType.MULTIPLE_CHOICE
        ):
            choices = [
                choice.strip() for choice in field_0.field_value.split(",")
            ]
            if all(
                choice in choice_grammar_element_field.options
                for choice in choices
            ):
                field_0.field_value = ", ".join(choices)
            else:
                self.add_error(
                    grammar_element_field.title,
                    (
                        f"Node's {grammar_element_field.title} must not contain"
                        f" values other than {', '.join(choice_grammar_element_field.options)}."
                    ),
                )



"""
@relation(SDOC-SRS-137, SDOC-SRS-140, scope=file)
"""

from copy import copy
from enum import Enum
from typing import Any, List, Optional

from strictdoc.backend.sdoc_source_code.models.requirement_marker import Req
from strictdoc.helpers.auto_described import auto_described


class RangeMarkerType(Enum):
    FUNCTION = "function"
    CLASS = "class"
    FILE = "file"


@auto_described
class FunctionRangeMarker:
    def __init__(
        self, parent: Any, reqs_objs: List[Req], scope: str, role: Optional[str]
    ):
        assert isinstance(reqs_objs, list)
        assert isinstance(scope, str), scope
        self.parent = parent
        self.reqs_objs: List[Req] = reqs_objs
        self.reqs: List[str] = list(map(lambda req: req.uid, reqs_objs))
        self.role: Optional[str] = (
            role if role is not None and len(role) > 0 else None
        )

        self.scope: RangeMarkerType = RangeMarkerType(scope)

        # Line number of the marker in the source code.
        self.ng_source_line_begin: Optional[int] = None
        self.ng_source_column_begin: Optional[int] = None

        # Line number of the marker range in the source code:
        # TODO: Improve description.
        # For Begin ranges:
        #   ng_range_line_begin == ng_source_line_begin  # noqa: ERA001
        #   ng_range_line_end == ng_source_line_begin of the End marker  # noqa: ERA001, E501
        # For End ranges:
        #   ng_range_line_begin == ng_range_line_begin of the Begin marker  # noqa: ERA001, E501
        #   ng_range_line_end == ng_source_line_begin  # noqa: ERA001
        self.ng_range_line_begin: Optional[int] = None
        self.ng_range_line_end: Optional[int] = None

        self.ng_marker_line: Optional[int] = None
        self.ng_marker_column: Optional[int] = None

        self.ng_is_nodoc = "nosdoc" in self.reqs

        self.begin_or_end: bool = True

        self._description: Optional[str] = None

    def is_forward(self) -> bool:
        return False

    def is_range_marker(self) -> bool:
        return True

    def is_line_marker(self) -> bool:
        return False

    def is_begin(self) -> bool:
        return self.begin_or_end

    def is_end(self) -> bool:
        return not self.begin_or_end

    def get_description(self) -> Optional[str]:
        return self._description

    def set_description(self, description: str) -> None:
        assert isinstance(description, str)
        self._description = description

    def create_end_marker(self) -> "FunctionRangeMarker":
        marker_copy = copy(self)
        marker_copy.begin_or_end = False
        marker_copy.ng_range_line_begin = self.ng_range_line_begin
        marker_copy.ng_source_line_begin = self.ng_range_line_end
        return marker_copy


@auto_described
class ForwardFunctionRangeMarker(FunctionRangeMarker):
    def __init__(
        self,
        parent: Any,
        reqs_objs: List[Req],
        scope: str,
        role: Optional[str] = None,
    ):
        super().__init__(parent, reqs_objs, scope, role)

    def is_forward(self) -> bool:
        return True



from strictdoc.core.file_tree import File, Folder
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex


def screen_should_display_folder(
    folder: Folder,
    traceability_index: TraceabilityIndex,
    project_config: ProjectConfig,
    must_only_include_non_included_sdoc: bool,
) -> bool:
    """
    Control if a folder appears on Project Index or Document screen.

    must_only_include_non_included_sdoc:
        see screen_should_display_file() for a description of the purpose.
    """
    assert isinstance(folder, Folder), folder

    if not folder.has_sdoc_content:
        return False

    for file_ in folder.files:
        if screen_should_display_file(
            file_,
            traceability_index,
            project_config,
            must_only_include_non_included_sdoc,
        ):
            return True

    for folder_ in folder.subfolder_trees:
        if screen_should_display_folder(
            folder_,
            traceability_index,
            project_config,
            must_only_include_non_included_sdoc,
        ):
            return True

    return False


def screen_should_display_file(
    file: File,
    traceability_index: TraceabilityIndex,
    project_config: ProjectConfig,
    must_only_include_non_included_sdoc: bool,
) -> bool:
    """
    Control if a file appears on Project Index or Document screen.

    must_only_include_non_included_sdoc:
        the argument controls whether to only display non-included SDoc
        documents or also included (fragment) documents. The Project Tree screen
        uses this as False while Document screen uses it as False where the
        standalone fragments are not printed, and instead the including documents
        prints its included documents recursively.
    """
    assert isinstance(file, File), file

    if (
        file.has_extension(".junit.xml")
        or file.has_extension(".gcov.json")
        or file.has_extension(".robot.xml")
        or file.has_extension(".reqif")
    ):
        return True

    if file.has_extension(".sdoc"):
        document = traceability_index.document_tree.get_document_by_path(
            file.full_path
        )
        if not document.document_is_included():
            return True

        if (
            not must_only_include_non_included_sdoc
            and project_config.export_included_documents
        ):
            return True

    return False



"""
@relation(SDOC-SRS-149, scope=file)
"""

from typing import Any

from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.mid import MID


@auto_described()
class InlineLink:
    def __init__(self, parent: Any, value: str) -> None:
        self.parent: Any = parent
        self.link: str = value

        # FIXME: Remove either mid or reserved_mid.
        self.mid: MID = MID.create()
        self.reserved_mid: MID = self.mid

    def parent_node(self) -> Any:
        return self.parent.parent



STRICTDOC_BASIC_TYPE_SYSTEM = r"""
BooleanChoice[noskipws]:
  ('True' | 'False')
;

ChoiceOption[noskipws]:
  /[\w\/|-]+( *[\w\/|-]+)*/
;

ChoiceOptionXs[noskipws]:
  /, /- ChoiceOption
;

RequirementType[noskipws]:
  !ReservedKeyword /[A-Z]+(_[A-Z]+)*/
;

ReservedKeyword[noskipws]:
  'DOCUMENT' | 'GRAMMAR'
;

Reference[noskipws]:
  // FileReference is an early, experimental feature. Do not use yet.
  ParentReqReference | ChildReqReference | FileReference
;

ParentReqReference[noskipws]:
  '- TYPE: Parent' '\n'
  '  VALUE: ' ref_uid = /.*$/ '\n'
  ('  ROLE: ' role = /.+$/ '\n')?
;

ChildReqReference[noskipws]:
  '- TYPE: Child' '\n'
  '  VALUE: ' ref_uid = /.*$/ '\n'
  ('  ROLE: ' role = /.+$/ '\n')?
;


FileReference[noskipws]:
  // FileReference is an early, experimental feature. Do not use yet.
  '- TYPE: File' '\n'
  ('  ROLE: ' role = /.+$/ '\n')?
  g_file_entry = FileEntry
;

FileEntry[noskipws]:
  ('  FORMAT: ' g_file_format = FileEntryFormat '\n')?
   '  VALUE: ' g_file_path = /.*$/ '\n'
  ('  LINE_RANGE: ' g_line_range = /.*$/ '\n')?
  ('  FUNCTION: ' function = /.*$/ '\n')?
  ('  CLASS: ' clazz = /.*$/ '\n')?
;

FileEntryFormat[noskipws]:
  'Sourcecode' | 'Python' | /[A-Z]+[A-Z_]*/
;

"""



"""
@relation(SDOC-SRS-152, scope=file)
"""

import os
from typing import List, NamedTuple, Optional, Tuple, Union

from strictdoc.backend.excel.import_.excel_sheet_proxy import ExcelSheetProxy
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import DocumentConfig
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementFieldString,
)
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.object_factory import SDocObjectFactory
from strictdoc.backend.sdoc.models.reference import ParentReqReference
from strictdoc.helpers.string import ensure_newline


class RequirementColumns(NamedTuple):
    uid_column: Optional[int]
    title_column: Optional[int]
    statement_column: Optional[int]
    comment_column: Optional[int]
    parent_column: Optional[int]
    extra_header_pairs: List[Tuple[int, str]]


def safe_name(dangerous_name: str) -> str:
    dangerous_name = dangerous_name.splitlines()[0]
    dangerous_name = dangerous_name.strip()
    dangerous_name = dangerous_name.upper()
    dangerous_name = dangerous_name.replace(":", "")
    dangerous_name = dangerous_name.replace("+", "")
    dangerous_name = dangerous_name.replace(",", "_")
    dangerous_name = dangerous_name.replace(" ", "_")
    dangerous_name = dangerous_name.replace("-", "_")
    dangerous_name = dangerous_name.replace("/", "_OR_")
    return dangerous_name


class ExcelToSDocConverter:
    @staticmethod
    def convert(
        excel_file: str, title: Union[str, None] = None
    ) -> SDocDocument:
        """
        Convert a provided Excel file to an SDoc document.

        Optional argument title is present for external scripts to assign a title.
        """
        sheet = ExcelSheetProxy(excel_file)

        excel_file_name = os.path.basename(excel_file)
        if title is None:
            title = excel_file_name + " sheet " + sheet.name

        all_header_columns = list(range(sheet.ncols))

        # The first 16 rows should do ¯\_(ツ)_/¯.
        for i in range(16):
            if sheet.row_values(i)[0].strip() != "":
                header_row_idx = i
                break
        else:
            raise NotImplementedError

        header_row = sheet.row_values(header_row_idx)
        safe_header_row = [safe_name(x) for x in header_row]

        statement_column = None
        uid_column = None
        comment_column = None
        title_column = None
        parent_column = None
        for header_column_idx, header_column in enumerate(safe_header_row):
            if header_column in ("REQUIREMENT", "STATEMENT"):
                statement_column = header_column_idx
            elif header_column in (
                "REF",
                "REF #",
                "REF_#",
                "REFDES",
                "ID",
                "UID",
            ):
                uid_column = header_column_idx
            elif header_column in ("REMARKS", "COMMENT"):
                comment_column = header_column_idx
            elif header_column in ("TITLE", "NAME"):
                title_column = header_column_idx
            elif header_column in ("PARENT", "PARENT_REF", "PARENT_UID"):
                parent_column = header_column_idx
            else:
                continue
            all_header_columns.remove(header_column_idx)
        assert statement_column is not None

        extra_header_pairs: List[Tuple[int, str]] = list(
            map(lambda x: (x, safe_header_row[x]), all_header_columns)
        )
        columns = RequirementColumns(
            uid_column=uid_column,
            title_column=title_column,
            statement_column=statement_column,
            comment_column=comment_column,
            parent_column=parent_column,
            extra_header_pairs=extra_header_pairs,
        )

        document = ExcelToSDocConverter.create_document(
            title, extra_header_pairs
        )
        for i in range(header_row_idx + 1, sheet.nrows):
            row_values = sheet.row_values(i)
            requirement = ExcelToSDocConverter.create_requirement(
                row_values, document, columns
            )
            document.section_contents.append(requirement)

        return document

    @staticmethod
    def create_document(
        title: Optional[str], extra_header_pairs: List[Tuple[int, str]]
    ) -> SDocDocument:
        document_config = DocumentConfig.default_config(None)
        document_title = title if title else "<No title>"
        document = SDocDocument(
            mid=None,
            title=document_title,
            config=document_config,
            view=None,
            grammar=None,
            section_contents=[],
        )

        # FIXME: This is becoming very limiting. It must work against a complete grammar.
        fields = list(
            DocumentGrammar.create_default(document)
            .elements_by_type["REQUIREMENT"]
            .fields
        )

        for _, name in extra_header_pairs:
            fields.extend(
                [
                    GrammarElementFieldString(
                        parent=None,
                        title=name,
                        human_title=None,
                        required="False",
                    ),
                ]
            )

        requirements_element = GrammarElement(
            parent=None,
            tag="REQUIREMENT",
            property_is_composite="",
            property_prefix="",
            property_view_style="",
            fields=fields,
            relations=[],
        )
        elements = [requirements_element]
        grammar = DocumentGrammar(parent=document, elements=elements)
        document.grammar = grammar
        return document

    @staticmethod
    def create_requirement(
        row_values: List[str],
        document: SDocDocument,
        columns: RequirementColumns,
    ) -> SDocNode:
        assert columns.statement_column is not None
        statement = row_values[columns.statement_column]
        uid = None
        if columns.uid_column is not None:
            uid = row_values[columns.uid_column].strip()
        title = None
        if columns.title_column is not None:
            title = row_values[columns.title_column].strip()
        comments = None
        if columns.comment_column is not None:
            comment = row_values[columns.comment_column].strip()
            if comment in ("", "-"):
                comments = None
            else:
                comments = [ensure_newline(comment)]
        parent_uid = None
        if columns.parent_column is not None:
            parent_uid = row_values[columns.parent_column].strip()
            if len(parent_uid) == 0:
                parent_uid = None

        template_requirement = SDocObjectFactory.create_requirement(
            document,
            node_type="REQUIREMENT",
            title=title,
            uid=uid,
            level=None,
            statement=None,
            statement_multiline=ensure_newline(statement),
            rationale=None,
            rationale_multiline=None,
            tags=None,
            comments=comments,
        )
        for i, name in columns.extra_header_pairs:
            value = row_values[i].strip()
            if value != "":
                template_requirement.ordered_fields_lookup[name] = [
                    SDocNodeField.create_from_string(
                        parent=None,
                        field_name=name,
                        field_value=ensure_newline(value),
                        multiline=True,
                    )
                ]
        if parent_uid is not None:
            reference = ParentReqReference(
                template_requirement, parent_uid, role=None
            )
            template_requirement.relations = [reference]
        requirement = SDocNode(
            parent=template_requirement.parent,
            node_type=template_requirement.node_type,
            fields=list(template_requirement.enumerate_fields()),
            relations=template_requirement.relations,
        )

        return requirement



from typing import Any, List, Optional

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.grammar_element import GrammarElement
from strictdoc.backend.sdoc.models.model import (
    SDocExtendedElementIF,
)
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.helpers.cast import assert_cast


class Expression:
    pass


class StringExpression:
    def __init__(self, parent: Any, string: str):
        self.parent: Any = parent
        self.string: str = string


class NoneExpression:
    def __init__(self, parent: Any, _: str):
        self.parent: Any = parent


class NodeFieldExpression:
    def __init__(self, parent: Any, field_name: str):
        self.parent: Any = parent
        self.field_name = field_name


class NodeContainsExpression:
    def __init__(self, parent: Any, string: str):
        self.parent: Any = parent
        self.string: str = string


class NodeHasParentRequirementsExpression:
    def __init__(self, parent: Any, _: Any) -> None:
        self.parent: Any = parent


class NodeContainsAnyFreeTextExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeHasChildRequirementsExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsRequirementExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsSourceFileExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsSourceFileWithCompleteCoverageExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsSourceFileWithPartialCoverageExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsSourceFileWithNoCoverageExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsSectionExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class NodeIsRootExpression:
    def __init__(self, parent: Any, _: Any):
        self.parent: Any = parent


class EqualExpression:
    def __init__(self, parent: Any, lhs_expr: Expression, rhs_expr: Expression):
        self.parent: Any = parent
        self.lhs_expr: Expression = lhs_expr
        self.rhs_expr: Expression = rhs_expr


class AndExpression:
    def __init__(self, parent: Any, expressions: List[Expression]):
        self.parent: Any = parent
        self.expressions: List[Expression] = expressions


class OrExpression:
    def __init__(self, parent: Any, expressions: List[Expression]):
        self.parent: Any = parent
        self.expressions: List[Expression] = expressions


class NotExpression:
    def __init__(self, parent: Any, expression: Expression):
        self.parent: Any = parent
        self.expression: Expression = expression


class NotEqualExpression:
    def __init__(self, parent: Any, lhs_expr: Expression, rhs_expr: Expression):
        self.parent: Any = parent
        self.lhs_expr: Expression = lhs_expr
        self.rhs_expr: Expression = rhs_expr


class InExpression:
    def __init__(self, parent: Any, lhs_expr: Expression, rhs_expr: Expression):
        self.parent: Any = parent
        self.lhs_expr: Expression = lhs_expr
        self.rhs_expr: Expression = rhs_expr


class NotInExpression:
    def __init__(self, parent: Any, lhs_expr: Expression, rhs_expr: Expression):
        self.parent: Any = parent
        self.lhs_expr: Expression = lhs_expr
        self.rhs_expr: Expression = rhs_expr


class Query:
    def __init__(self, root_expression: Any):
        self.root_expression: Any = root_expression


class QueryNullObject:
    def evaluate(self, _: Any) -> bool:
        return True


class QueryObject:
    def __init__(
        self, query: Query, traceability_index: TraceabilityIndex
    ) -> None:
        self.query: Query = query
        self.traceability_index: TraceabilityIndex = traceability_index

    def evaluate(self, node: SDocExtendedElementIF) -> bool:
        return self._evaluate(node, self.query.root_expression)

    def _evaluate(self, node: SDocExtendedElementIF, expression: Any) -> bool:
        if isinstance(expression, EqualExpression):
            return self._evaluate_equal(node, expression)
        if isinstance(expression, NotEqualExpression):
            return self._evaluate_not_equal(node, expression)
        if isinstance(expression, NodeContainsExpression):
            return self._evaluate_node_contains(node, expression)
        if isinstance(expression, NodeContainsAnyFreeTextExpression):
            return self._evaluate_node_contains_any_text(node)
        if isinstance(expression, NodeHasParentRequirementsExpression):
            return self._evaluate_node_has_parent_requirements(node)
        if isinstance(expression, NodeHasChildRequirementsExpression):
            return self._evaluate_node_has_child_requirements(node)
        if isinstance(expression, NodeIsRequirementExpression):
            return (
                isinstance(node, SDocNode) and node.node_type == "REQUIREMENT"
            )
        if isinstance(expression, NodeIsSectionExpression):
            return (
                isinstance(node, SDocNode) and node.node_type == "SECTION"
            ) or isinstance(node, SDocSection)
        if isinstance(expression, NodeIsSourceFileExpression):
            return isinstance(node, SourceFileTraceabilityInfo)
        if isinstance(
            expression, NodeIsSourceFileWithCompleteCoverageExpression
        ):
            return (
                isinstance(node, SourceFileTraceabilityInfo)
                and node.get_coverage() == 100
            )
        if isinstance(
            expression, NodeIsSourceFileWithPartialCoverageExpression
        ):
            return isinstance(node, SourceFileTraceabilityInfo) and (
                0 < node.get_coverage() < 100
            )
        if isinstance(expression, NodeIsSourceFileWithNoCoverageExpression):
            return (
                isinstance(node, SourceFileTraceabilityInfo)
                and node.get_coverage() == 0
            )
        if isinstance(expression, NodeIsRootExpression):
            if isinstance(node, SDocNode):
                return node.is_root
            raise RuntimeError(
                "The node.is_root expression can be only called on nodes."
            )
        if isinstance(expression, NotExpression):
            return not self._evaluate(node, expression.expression)
        if isinstance(expression, AndExpression):
            for sub_expression_ in expression.expressions:
                if not self._evaluate(node, sub_expression_):
                    return False
            return True
        if isinstance(expression, OrExpression):
            for sub_expression_ in expression.expressions:
                if self._evaluate(node, sub_expression_):
                    return True
            return False
        if isinstance(expression, InExpression):
            if (
                rhs_value := self._evaluate_value(node, expression.rhs_expr)
            ) is not None and (
                lhs_value := self._evaluate_value(node, expression.lhs_expr)
            ) is not None:
                return lhs_value in rhs_value
            return False
        if isinstance(expression, NotInExpression):
            if (
                rhs_value := self._evaluate_value(node, expression.rhs_expr)
            ) is not None and (
                lhs_value := self._evaluate_value(node, expression.lhs_expr)
            ) is not None:
                return lhs_value not in rhs_value
            return False
        assert 0, expression

    def _evaluate_equal(
        self, node: SDocExtendedElementIF, expression: EqualExpression
    ) -> bool:
        return self._evaluate_value(
            node, expression.lhs_expr
        ) == self._evaluate_value(node, expression.rhs_expr)

    def _evaluate_not_equal(
        self, node: SDocExtendedElementIF, expression: NotEqualExpression
    ) -> bool:
        return self._evaluate_value(
            node, expression.lhs_expr
        ) != self._evaluate_value(node, expression.rhs_expr)

    def _evaluate_value(
        self, node: SDocExtendedElementIF, expression: Any
    ) -> Optional[str]:
        if isinstance(expression, NodeFieldExpression):
            return self._evaluate_node_field_expression(node, expression)
        if isinstance(expression, StringExpression):
            return expression.string
        if isinstance(expression, NoneExpression):
            return None
        assert 0, expression

    def _evaluate_node_field_expression(
        self,
        node: SDocExtendedElementIF,
        expression: NodeFieldExpression,
    ) -> Optional[str]:
        field_name = expression.field_name
        if (
            isinstance(node, SDocNode)
            and node.is_requirement()
            and node.node_type == "REQUIREMENT"
        ):
            requirement: SDocNode = assert_cast(node, SDocNode)
            requirement_document: SDocDocument = assert_cast(
                requirement.get_document(), SDocDocument
            )
            document_grammar: DocumentGrammar = assert_cast(
                requirement_document.grammar, DocumentGrammar
            )
            element: GrammarElement = document_grammar.elements_by_type[
                requirement.node_type
            ]
            grammar_field_titles = list(map(lambda f: f.title, element.fields))
            if field_name not in grammar_field_titles:
                return None
            field_value = requirement._get_cached_field(field_name, False)
            if field_value is not None:
                return field_value
            return None
        elif isinstance(node, SDocSection) and node.is_section():
            section: SDocSection = assert_cast(node, SDocSection)
            if field_name == "UID":
                return section.reserved_uid
            elif field_name == "TITLE":
                return section.title
            raise AttributeError(f"No such section field: {field_name}.")
        else:
            raise NotImplementedError

    def _evaluate_node_has_parent_requirements(
        self, node: SDocExtendedElementIF
    ) -> bool:
        if not isinstance(node, SDocNode):
            raise TypeError(
                f"node.has_parent_requirements can be only called on "
                f"Requirement objects, got: {node.__class__.__name__}. To fix "
                f"the error, prepend your query with node.is_requirement."
            )
        return self.traceability_index.has_parent_requirements(node)

    def _evaluate_node_has_child_requirements(
        self, node: SDocExtendedElementIF
    ) -> bool:
        if not isinstance(node, SDocNode):
            raise TypeError(
                f"node.has_child_requirements can be only called on "
                f"Requirement objects, got: {node.__class__.__name__}. To fix "
                f"the error, prepend your query with node.is_requirement."
            )
        return self.traceability_index.has_children_requirements(node)

    def _evaluate_node_contains(
        self,
        node: SDocExtendedElementIF,
        expression: NodeContainsExpression,
    ) -> bool:
        if isinstance(node, SDocNode):
            requirement = assert_cast(node, SDocNode)
            requirement_field_: SDocNodeField
            for requirement_field_ in requirement.enumerate_fields():
                if expression.string in requirement_field_.get_text_value():
                    return True
            return False
        if isinstance(node, SDocSection):
            section = assert_cast(node, SDocSection)
            if expression.string in section.title:
                return True
            return False
        raise NotImplementedError

    def _evaluate_node_contains_any_text(
        self, node: SDocExtendedElementIF
    ) -> bool:
        if not isinstance(node, SDocSection) and not (
            isinstance(node, SDocNode) and node.node_type == "SECTION"
        ):
            raise TypeError(
                f"node.contains_any_text can be only called on "
                f"SECTION objects, got: {node.__class__.__name__}. To fix "
                f"the error, prepend your query with node.is_section."
            )
        return node.has_any_text_nodes()



"""
@relation(SDOC-SRS-109, scope=file)
"""

from typing import Generator, List, Optional, Union

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentFromFileIF,
    SDocDocumentIF,
    SDocElementIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import SDocCompositeNode
from strictdoc.helpers.auto_described import auto_described


@auto_described
class DocumentFromFile(SDocDocumentFromFileIF):
    def __init__(
        self,
        parent: Union[SDocDocumentIF, SDocSectionIF],
        file: str,
    ) -> None:
        self.parent: Union[SDocDocumentIF, SDocSectionIF] = parent
        self.file: str = file

        self.ng_has_requirements: bool = False
        self.ng_document_reference: Optional[DocumentReference] = None
        self.ng_included_document_reference: Optional[DocumentReference] = None
        self.ng_resolved_custom_level: Optional[str] = None
        self.ng_whitelisted: bool = True
        self.resolved_full_path_to_document_file: Optional[str] = None
        self.resolved_document: Optional[SDocDocumentIF] = None

    def iterate_nodes(
        self, element_type: Optional[str] = None
    ) -> Generator[SDocNodeIF, None, None]:
        """
        Iterate over all non-[TEXT] nodes in the document.

        If element_type is given, then only nodes of type `element_type` are
        returned. Otherwise, all element types are returned.
        """

        task_list: List[SDocElementIF] = list(self.section_contents)
        while task_list:
            node = task_list.pop(0)

            if isinstance(node, SDocDocumentFromFileIF):
                yield from node.iterate_nodes(element_type)

            if isinstance(node, SDocNodeIF):
                if node.node_type == "TEXT":
                    continue
                if element_type is None or node.node_type == element_type:
                    yield node

            task_list.extend(node.section_contents)

    def has_any_requirements(self) -> bool:
        return any(True for _ in self.iterate_nodes())

    @property
    def section_contents(self) -> List[SDocDocumentIF]:
        assert self.resolved_document is not None
        return [self.resolved_document]

    def configure_with_resolved_document(
        self, resolved_document: SDocDocument
    ) -> None:
        assert self.ng_included_document_reference is None
        assert self.ng_document_reference is not None

        assert resolved_document is not None
        self.resolved_document = resolved_document

        including_document_or_section = self.parent

        if isinstance(including_document_or_section, SDocDocumentIF):
            including_document = including_document_or_section
        elif isinstance(
            including_document_or_section, (SDocSectionIF, SDocCompositeNode)
        ):
            including_document_: Optional[SDocDocumentIF] = (
                including_document_or_section.get_document()
            )
            assert isinstance(including_document_, SDocDocumentIF)
            including_document = including_document_
        else:
            raise AssertionError(including_document_or_section)

        resolved_document.ng_including_document_from_file = self

        assert resolved_document.ng_including_document_reference is not None
        resolved_document.ng_including_document_reference.set_document(
            including_document
        )

        including_document.included_documents.append(resolved_document)



"""
@relation(SDOC-SRS-18, scope=file)
"""

from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_config import (
    DocumentConfig,
    DocumentCustomMetadata,
    DocumentCustomMetadataKeyValuePair,
)
from strictdoc.backend.sdoc.models.document_from_file import DocumentFromFile
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
)
from strictdoc.backend.sdoc.models.document_view import (
    DocumentView,
    ViewElement,
    ViewElementField,
    ViewElementHiddenTag,
    ViewElementTags,
)
from strictdoc.backend.sdoc.models.grammar_element import (
    GrammarElement,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldString,
    GrammarElementFieldTag,
    GrammarElementRelationChild,
    GrammarElementRelationFile,
    GrammarElementRelationParent,
)
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.node import (
    SDocCompositeNode,
    SDocNode,
    SDocNodeField,
)
from strictdoc.backend.sdoc.models.reference import (
    ChildReqReference,
    FileEntry,
    FileReference,
    ParentReqReference,
    Reference,
)
from strictdoc.backend.sdoc.models.section import SDocSection

SECTION_MODELS = [
    SDocSection,
    DocumentFromFile,
    SDocNode,
    SDocNodeField,
    SDocCompositeNode,
    Reference,
    ParentReqReference,
    ChildReqReference,
    FileReference,
    InlineLink,
    Anchor,
    FileEntry,
]

GRAMMAR_MODELS = [
    DocumentGrammar,
    GrammarElement,
    GrammarElementFieldString,
    GrammarElementFieldSingleChoice,
    GrammarElementFieldMultipleChoice,
    GrammarElementFieldTag,
    GrammarElementRelationParent,
    GrammarElementRelationChild,
    GrammarElementRelationFile,
]

DOCUMENT_MODELS = [
    DocumentConfig,
    DocumentCustomMetadata,
    DocumentCustomMetadataKeyValuePair,
    SDocDocument,
    DocumentView,
    ViewElement,
    ViewElementField,
    ViewElementTags,
    ViewElementHiddenTag,
]
DOCUMENT_MODELS.extend(GRAMMAR_MODELS)
DOCUMENT_MODELS.extend(SECTION_MODELS)



import importlib
import os
import sys
from collections import defaultdict
from functools import partial
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import orjson
from html2pdf4doc.html2pdf4doc import PATH_TO_HTML2PDF4DOC_JS

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.core.asset_manager import AssetDir
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.core.project_config import ProjectConfig, ProjectFeature
from strictdoc.core.source_tree import SourceTree
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.document_type import DocumentType
from strictdoc.export.html.generators.document import DocumentHTMLGenerator
from strictdoc.export.html.generators.document_deep_trace import (
    DocumentDeepTraceHTMLGenerator,
)
from strictdoc.export.html.generators.document_pdf import (
    DocumentHTML2PDFGenerator,
)
from strictdoc.export.html.generators.document_table import (
    DocumentTableHTMLGenerator,
)
from strictdoc.export.html.generators.document_trace import (
    DocumentTraceHTMLGenerator,
)
from strictdoc.export.html.generators.document_tree import (
    DocumentTreeHTMLGenerator,
)
from strictdoc.export.html.generators.project_map import (
    ProjectMapGenerator,
)
from strictdoc.export.html.generators.project_statistics import (
    ProgressStatisticsGenerator,
)
from strictdoc.export.html.generators.source_file_coverage import (
    SourceFileCoverageHTMLGenerator,
)
from strictdoc.export.html.generators.source_file_view_generator import (
    SourceFileViewHTMLGenerator,
)
from strictdoc.export.html.generators.traceability_matrix import (
    TraceabilityMatrixHTMLGenerator,
)
from strictdoc.export.html.html_templates import HTMLTemplates
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.export.html.renderers.markup_renderer import MarkupRenderer
from strictdoc.export.html.tools.html_embedded import HTMLEmbedder
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.exception import StrictDocException
from strictdoc.helpers.file_modification_time import get_file_modification_time
from strictdoc.helpers.file_system import sync_dir
from strictdoc.helpers.git_client import GitClient
from strictdoc.helpers.mid import MID
from strictdoc.helpers.parallelizer import Parallelizer
from strictdoc.helpers.paths import SDocRelativePath
from strictdoc.helpers.timing import measure_performance, timing_decorator


class HTMLGenerator:
    def __init__(
        self, project_config: ProjectConfig, html_templates: HTMLTemplates
    ):
        self.project_config: ProjectConfig = project_config
        self.html_templates = html_templates
        self.git_client: GitClient = GitClient()

    def export_complete_tree(
        self,
        *,
        traceability_index: TraceabilityIndex,
        parallelizer: Parallelizer,
    ) -> None:
        Path(self.project_config.export_output_html_root).mkdir(
            parents=True, exist_ok=True
        )

        # Export assets.
        HTMLGenerator.export_assets(
            traceability_index=traceability_index,
            project_config=self.project_config,
            export_output_html_root=self.project_config.export_output_html_root,
        )

        # Export static search index.
        self.export_static_html_search_index(
            traceability_index=traceability_index
        )

        # Export all documents in parallel.
        export_binding = partial(
            self.export_single_document_with_performance,
            traceability_index=traceability_index,
        )

        # By default, do not export included documents. Only, if the option to
        # include is provided.
        documents_to_export: List[SDocDocument] = []

        if self.project_config.export_included_documents:
            documents_to_export[:] = (
                traceability_index.document_tree.document_list
            )
        else:
            for document_ in traceability_index.document_tree.document_list:
                if document_.document_is_included():
                    continue

                document_meta = assert_cast(document_.meta, DocumentMeta)

                input_doc_full_path = document_meta.input_doc_full_path
                output_doc_full_path = document_meta.output_document_full_path

                if os.path.isfile(output_doc_full_path) and (
                    get_file_modification_time(input_doc_full_path)
                    < get_file_modification_time(output_doc_full_path)
                    and not traceability_index.file_dependency_manager.must_generate(
                        document_meta.output_document_full_path
                    )
                ):
                    with measure_performance(f"Skip: {document_.title}"):
                        continue

                documents_to_export.append(document_)

        if len(documents_to_export) > 0:
            if len(traceability_index.document_tree.document_list) <= 25:
                parallelizer.run_parallel(documents_to_export, export_binding)
            else:
                print(  # noqa: T201
                    "NOTE: Running document export without parallelization "
                    "because the document tree contains more than 25 documents."
                )
                for document_ in documents_to_export:
                    export_binding(document_)

        # Export document tree.
        # FIXME: It is important that this export is **after** the parallelized
        # export of single documents. It turns out that Jinja does not play
        # well with the multiprocessing's processed-based parallelization.
        # _pickle.PicklingError: Can't pickle <function sync_do_first at 0x1077bdf80>: it's not the same object as jinja2.filters.sync_do_first.
        self.export_project_tree_screen(traceability_index=traceability_index)

        # Export JavaScript map of the document tree (project map)
        self.export_project_map(traceability_index=traceability_index)

        # Export project statistics.
        if self.project_config.is_feature_activated(
            ProjectFeature.PROJECT_STATISTICS_SCREEN
        ):
            self.export_project_statistics(traceability_index)

        # Export requirements coverage.
        if self.project_config.is_feature_activated(
            ProjectFeature.TRACEABILITY_MATRIX_SCREEN
        ):
            self.export_requirements_coverage_screen(
                traceability_index=traceability_index,
            )

        # Export source coverage.
        if self.project_config.is_feature_activated(
            ProjectFeature.REQUIREMENT_TO_SOURCE_TRACEABILITY
        ):
            self.export_source_coverage_screen(
                traceability_index=traceability_index,
            )

        print(  # noqa: T201
            "Export completed. Documentation tree can be found at:\n"
            f"{self.project_config.export_output_html_root}"
        )

    @staticmethod
    def export_assets(
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
        export_output_html_root: str,
        flat_assets: bool = False,
    ) -> None:
        """
        Copy all assets to output dir during HTML/PDF generation.

        :param bool flat_assets: This parameter is always set to False except when
                                 exporting a "bundle document" with HTML2PDF.
                                 The bundle document contains all documents of
                                 the documentation tree. In this case, all assets
                                 are simply copied to the top level _assets folder,
                                 independently on how nested the contained documents are.
        """

        # Export StrictDoc's own assets.
        output_html_static_files = os.path.join(
            export_output_html_root,
            project_config.dir_for_sdoc_assets,
        )
        sync_dir(
            project_config.get_static_files_path(),
            output_html_static_files,
            message="Copying StrictDoc's assets",
        )

        # Export MathJax.
        if project_config.is_feature_activated(ProjectFeature.MATHJAX):
            output_html_mathjax = os.path.join(
                export_output_html_root,
                project_config.dir_for_sdoc_assets,
                "mathjax",
            )
            Path(output_html_mathjax).mkdir(parents=True, exist_ok=True)
            mathjax_src = os.path.join(
                project_config.get_extra_static_files_path(), "mathjax"
            )
            sync_dir(
                mathjax_src,
                output_html_mathjax,
                message="Copying MathJax assets",
            )

        # Export Mermaid.
        if project_config.is_feature_activated(ProjectFeature.MERMAID):
            output_html_mathjax = os.path.join(
                export_output_html_root,
                project_config.dir_for_sdoc_assets,
                "mermaid",
            )
            Path(output_html_mathjax).mkdir(parents=True, exist_ok=True)
            mermaid_src = os.path.join(
                project_config.get_extra_static_files_path(), "mermaid"
            )
            sync_dir(
                mermaid_src,
                output_html_mathjax,
                message="Copying Mermaid assets",
            )

        # Export Rapidoc.
        if project_config.is_feature_activated(ProjectFeature.RAPIDOC):
            output_html_rapidoc = os.path.join(
                export_output_html_root,
                project_config.dir_for_sdoc_assets,
                "rapidoc",
            )
            Path(output_html_rapidoc).mkdir(parents=True, exist_ok=True)
            rapidoc_src = os.path.join(
                project_config.get_extra_static_files_path(), "rapidoc"
            )
            sync_dir(
                rapidoc_src,
                output_html_rapidoc,
                message="Copying Rapidoc assets",
            )

        # Export NESTOR.
        if project_config.is_feature_activated(ProjectFeature.NESTOR):
            output_html_nestor = os.path.join(
                export_output_html_root,
                project_config.dir_for_sdoc_assets,
                "nestor",
            )
            Path(output_html_nestor).mkdir(parents=True, exist_ok=True)
            mathjax_src = os.path.join(
                project_config.get_extra_static_files_path(), "nestor"
            )
            sync_dir(
                mathjax_src,
                output_html_nestor,
                message="Copying Nestor assets",
            )

        # Export HTML2PDF.
        if project_config.is_feature_activated(ProjectFeature.HTML2PDF):
            sync_dir(
                os.path.dirname(PATH_TO_HTML2PDF4DOC_JS),
                output_html_static_files,
                message="Copying HTML2PDF.js",
            )

        # Export custom html2pdf template.
        if project_config.html2pdf_template is not None:
            output_custom_html2pdf_template = os.path.join(
                export_output_html_root,
                project_config.dir_for_sdoc_assets,
                "html2pdf_template",
            )
            sync_dir(
                os.path.abspath(
                    os.path.dirname(project_config.html2pdf_template)
                ),
                output_custom_html2pdf_template,
                message="Copying Custom HTML2PDF template assets",
            )

        # Export project's assets.

        redundant_assets: Dict[str, List[SDocRelativePath]] = {}
        for document_ in traceability_index.document_tree.document_list:
            assert document_.meta is not None
            for (
                included_document_
            ) in document_.iterate_included_documents_depth_first():
                assert included_document_.meta is not None

                redundant_assets.setdefault(
                    document_.meta.input_doc_assets_dir_rel_path.relative_path_posix,
                    [],
                )
                redundant_assets[
                    document_.meta.input_doc_assets_dir_rel_path.relative_path_posix
                ].append(included_document_.meta.input_doc_assets_dir_rel_path)

        assert traceability_index.asset_manager is not None

        asset_dir_: AssetDir
        for asset_dir_ in traceability_index.asset_manager.iterate():
            source_path = asset_dir_.full_path
            output_relative_path = asset_dir_.relative_path

            destination_path = os.path.join(
                export_output_html_root,
                output_relative_path.relative_path
                if not flat_assets
                else "_assets",
            )

            sync_dir(
                source_path,
                destination_path,
                message=f'Copying project assets "{output_relative_path.relative_path}"',
            )
            redundant_asset_paths = redundant_assets.get(
                output_relative_path.relative_path_posix
            )
            if redundant_asset_paths is not None:
                for redundant_asset_ in redundant_asset_paths:
                    destination_path = os.path.join(
                        export_output_html_root,
                        redundant_asset_.relative_path
                        if not flat_assets
                        else "_assets",
                    )
                    sync_dir(
                        source_path,
                        destination_path,
                        message=f'Copying project assets "{output_relative_path.relative_path}"',
                    )

    def export_single_document_with_performance(
        self,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
        specific_documents: Optional[Tuple[DocumentType, ...]] = None,
    ) -> None:
        if specific_documents is None:
            specific_documents = DocumentType.all()

        with measure_performance(f"Published: {document.title}"):
            self.export_single_document(
                document,
                traceability_index,
                specific_documents=specific_documents,
            )

    def export_single_document(
        self,
        document: SDocDocument,
        traceability_index: TraceabilityIndex,
        specific_documents: Optional[Tuple[DocumentType, ...]] = None,
    ) -> SDocDocument:
        if document.config.layout == "Website":
            specific_documents = (DocumentType.DOCUMENT,)
        elif specific_documents is None:
            specific_documents = DocumentType.all()

        assert document.meta is not None

        document_meta: DocumentMeta = document.meta

        document_output_folder = document_meta.output_document_dir_full_path
        Path(document_output_folder).mkdir(parents=True, exist_ok=True)

        root_path = document.meta.get_root_path_prefix()
        link_renderer = LinkRenderer(
            root_path=root_path,
            static_path=self.project_config.dir_for_sdoc_assets,
        )
        markup_renderer = MarkupRenderer.create(
            document.config.markup,
            traceability_index,
            link_renderer,
            self.html_templates,
            self.project_config,
            document,
        )

        if DocumentType.DOCUMENT in specific_documents:
            # Single Document pages.
            document_content = DocumentHTMLGenerator.export(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                standalone=False,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_doc_path()
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content)

        # Single Document Table pages.
        if (
            self.project_config.is_feature_activated(
                ProjectFeature.TABLE_SCREEN
            )
            and DocumentType.TABLE in specific_documents
        ):
            document_content = DocumentTableHTMLGenerator.export(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_table_path()
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content)

        # Single Document Traceability pages.
        if (
            self.project_config.is_feature_activated(
                ProjectFeature.TRACEABILITY_SCREEN
            )
            and DocumentType.TRACE in specific_documents
        ):
            document_content = DocumentTraceHTMLGenerator.export(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_traceability_path()
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content)

        # Single Document Deep Traceability pages.
        if (
            self.project_config.is_feature_activated(
                ProjectFeature.DEEP_TRACEABILITY_SCREEN
            )
            and DocumentType.DEEPTRACE in specific_documents
        ):
            document_content = DocumentDeepTraceHTMLGenerator.export_deep(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_deep_traceability_path()
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content)

        # Single Document PDF pages.
        if (
            self.project_config.is_feature_activated(ProjectFeature.HTML2PDF)
            and DocumentType.PDF in specific_documents
        ):
            document_content = DocumentHTML2PDFGenerator.export(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                standalone=False,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_pdf_path()
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content)

        if self.project_config.is_feature_activated(
            ProjectFeature.STANDALONE_DOCUMENT_SCREEN
        ):
            # Single Document pages (standalone)
            document_content = DocumentHTMLGenerator.export(
                self.project_config,
                document,
                traceability_index,
                markup_renderer,
                link_renderer,
                git_client=self.git_client,
                standalone=True,
                html_templates=self.html_templates,
            )
            document_out_file = document_meta.get_html_doc_standalone_path()
            document_content_with_embedded_assets = HTMLEmbedder.embed_assets(
                document_content, document_out_file
            )
            with open(document_out_file, "w", encoding="utf8") as file:
                file.write(document_content_with_embedded_assets)

        return document

    def export_project_tree_screen(
        self,
        *,
        traceability_index: TraceabilityIndex,
    ) -> None:
        Path(self.project_config.export_output_html_root).mkdir(
            parents=True, exist_ok=True
        )
        output_file = os.path.join(
            self.project_config.export_output_html_root, "index.html"
        )
        writer = DocumentTreeHTMLGenerator()
        output = writer.export(
            self.project_config,
            traceability_index=traceability_index,
            html_templates=self.html_templates,
        )
        with open(output_file, "w", encoding="utf8") as file:
            file.write(output)

    def export_project_map(
        self,
        *,
        traceability_index: TraceabilityIndex,
    ) -> None:
        assets_dir = os.path.join(
            self.project_config.export_output_html_root,
            self.project_config.dir_for_sdoc_assets,
        )
        output_file = os.path.join(assets_dir, "project_map.js")
        writer = ProjectMapGenerator()
        output = writer.export(
            self.project_config,
            traceability_index=traceability_index,
            html_templates=self.html_templates,
        )
        with open(output_file, "w", encoding="utf8") as file:
            file.write(output)

    def export_requirements_coverage_screen(
        self,
        *,
        traceability_index: TraceabilityIndex,
    ) -> None:
        requirements_coverage_content = TraceabilityMatrixHTMLGenerator.export(
            project_config=self.project_config,
            traceability_index=traceability_index,
            html_templates=self.html_templates,
        )
        output_html_requirements_coverage = os.path.join(
            self.project_config.export_output_html_root,
            "traceability_matrix.html",
        )
        with open(
            output_html_requirements_coverage, "w", encoding="utf8"
        ) as file:
            file.write(requirements_coverage_content)

    def export_source_coverage_screen(
        self,
        *,
        traceability_index: TraceabilityIndex,
    ) -> None:
        assert isinstance(
            traceability_index.document_tree.source_tree, SourceTree
        ), traceability_index.document_tree.source_tree
        print("Generating source files:")  # noqa: T201
        for (
            source_file
        ) in traceability_index.document_tree.source_tree.source_files:
            if not source_file.is_referenced:
                continue

            SourceFileViewHTMLGenerator.export_to_file(
                project_config=self.project_config,
                source_file=source_file,
                traceability_index=traceability_index,
                html_templates=self.html_templates,
            )

        source_coverage_content = SourceFileCoverageHTMLGenerator.export(
            project_config=self.project_config,
            traceability_index=traceability_index,
            html_templates=self.html_templates,
        )
        output_html_source_coverage = os.path.join(
            self.project_config.export_output_html_root, "source_coverage.html"
        )
        with open(output_html_source_coverage, "w", encoding="utf8") as file:
            file.write(source_coverage_content)

    def export_project_statistics(
        self,
        traceability_index: TraceabilityIndex,
    ) -> None:
        """
        Export project statistics to a dedicated HTML page.

        @relation(SDOC-SRS-97, scope=function)
        @relation(SDOC-SRS-154, scope=function)
        """

        link_renderer = LinkRenderer(
            root_path="",
            static_path=self.project_config.dir_for_sdoc_assets,
        )

        statistics_generator = ProgressStatisticsGenerator

        if (
            custom_statistics_generator_path
            := self.project_config.statistics_generator
        ) is not None:
            # It is important to add the input folder to the import path.
            # Otherwise, the custom statistics generator may not be found.
            # In fact, a more reasonable path to add would be the project config
            # path, but since it is not maintained by ProjectConfig yet and
            # usually equals the input path, add the input path for
            # now.
            input_paths = self.project_config.input_paths
            assert input_paths is not None and len(input_paths) > 0, (
                "Expected a valid input path."
            )
            sys.path.insert(0, input_paths[0])

            module_path, class_name = custom_statistics_generator_path.rsplit(
                ".", 1
            )
            try:
                module = importlib.import_module(module_path)
                statistics_generator = getattr(module, class_name)
            except ModuleNotFoundError as module_not_found_error_:
                raise StrictDocException(
                    "Could not import a user-provided statistics generator: "
                    f"{module_not_found_error_}."
                ) from module_not_found_error_

        document_content = statistics_generator.export(
            self.project_config,
            traceability_index,
            link_renderer,
            html_templates=self.html_templates,
        )
        output_html_source_coverage = os.path.join(
            self.project_config.export_output_html_root,
            "project_statistics.html",
        )
        with open(output_html_source_coverage, "w", encoding="utf8") as file:
            file.write(document_content)

    @timing_decorator("Export static HTML search index")
    def export_static_html_search_index(
        self,
        traceability_index: TraceabilityIndex,
    ) -> None:
        """
        Export a static search index as dictionaries in .js files.

        @relation(SDOC-SRS-155, scope=function)
        @relation(SDOC-SRS-156, scope=function)
        """

        # First check if there is nothing to do because no documents have been
        # changed or regenerated.
        for document_ in traceability_index.document_tree.document_list:
            assert document_.meta is not None
            if traceability_index.file_dependency_manager.must_generate(
                document_.meta.output_document_full_path
            ):
                break
        else:
            print(  # noqa: T201
                "All documents are up-to-date. "
                "Skipping the generation of a search index."
            )
            # If no documents need to be regenerated, set the search_index_timestamp
            # to the timestamp of the first document. The assumption here is
            # that StrictDoc does not randomize the document list, and the first
            # document will always be the same.
            # The HTML/JS code can rely on this timestamp to decide whether it
            # has to re-read the search index from the JS file or it can simply
            # fetch it from the DB which is 2x faster when it comes to very
            # large indexes.
            if len(traceability_index.document_tree.document_list) > 0:
                first_document = traceability_index.document_tree.document_list[
                    0
                ]
                assert first_document.meta is not None
                traceability_index.search_index_timestamp = (
                    get_file_modification_time(
                        first_document.meta.input_doc_full_path
                    )
                )
            return

        global_index: Dict[str, Set[int]] = defaultdict(set)
        global_map_nodes_by_mid: Dict[int, Dict[str, str]] = {}

        document_index_list: List[Dict[str, Set[str]]] = []
        document_map_list: List[Dict[int, Dict[str, str]]] = []

        map_mid_to_numbers: Dict[str, int] = {}

        with measure_performance("Build search index"):
            for document_ in traceability_index.document_tree.document_list:
                assert document_.meta is not None
                document_index_list.append(
                    document_.search_index.document_index
                )
                map_nodes_by_numbers: Dict[int, Dict[str, str]] = {}
                for (
                    node_mid_,
                    node_dict_,
                ) in document_.search_index.map_nodes_by_mid.items():
                    if node_mid_ not in map_mid_to_numbers:
                        map_mid_to_numbers[node_mid_] = (
                            len(map_mid_to_numbers) + 1
                        )
                    document_mid_number = map_mid_to_numbers[node_mid_]
                    assert isinstance(document_mid_number, int)
                    map_nodes_by_numbers[document_mid_number] = node_dict_

                document_map_list.append(map_nodes_by_numbers)
            for document_index_ in document_index_list:
                for term_, document_mids_ in document_index_.items():
                    document_mid_numbers = set()
                    for document_mid_ in document_mids_:
                        document_mid_number = map_mid_to_numbers[document_mid_]
                        document_mid_numbers.add(document_mid_number)
                    global_index[term_].update(document_mid_numbers)
            for map_nodes_by_mid_ in document_map_list:
                global_map_nodes_by_mid.update(map_nodes_by_mid_)

        link_renderer = LinkRenderer(
            root_path="",
            static_path=self.project_config.dir_for_sdoc_assets,
        )
        for _, node_ in global_map_nodes_by_mid.items():
            node = traceability_index.get_node_by_mid(MID(node_["MID"]))
            node_["_LINK"] = link_renderer.render_local_anchor(node)

        def default(obj: Any) -> Any:
            if isinstance(obj, set):
                return list(obj)
            raise TypeError

        with measure_performance("Serialize search index to JS"):
            document_content = (
                b"window.SDOC_SEARCH_INDEX = "
                + orjson.dumps(
                    global_index,
                    option=orjson.OPT_NON_STR_KEYS,
                    default=default,
                )
                + b";\n\n"
            )

        with measure_performance("Serialize lookup map {MID => node} to JS"):
            document_content += (
                b"window.SDOC_MAP_MID_TO_NODES = "
                + orjson.dumps(
                    global_map_nodes_by_mid, option=orjson.OPT_NON_STR_KEYS
                )
                + b";\n"
            )

        # Export StrictDoc's own assets.
        output_html_static_files = os.path.join(
            self.project_config.export_output_html_root,
            self.project_config.dir_for_sdoc_assets,
        )
        output_html_source_coverage = os.path.join(
            output_html_static_files,
            "static_html_search_index.js",
        )
        with open(output_html_source_coverage, "wb") as file:
            file.write(document_content)



import os
import subprocess
from typing import List, Optional


class GitClient:
    def __init__(self) -> None:
        self._commit_hash: Optional[str] = None
        self._commit_date: Optional[str] = None
        self._commit_datetime: Optional[str] = None
        self._branch: Optional[str] = None

    def get_commit_hash(self) -> str:
        if self._commit_hash is not None:
            return self._commit_hash
        commit_hash = self._run_git_command(
            ["git", "describe", "--always", "--tags"],
        )
        self._commit_hash = commit_hash.strip()
        return self._commit_hash

    def get_commit_date(self) -> str:
        if self._commit_date is not None:
            return self._commit_date
        commit_date = self._run_git_command(
            ["git", "log", "-1", "--format=%cd", "--date=format:%Y-%m-%d"]
        )
        self._commit_date = commit_date.strip()
        return self._commit_date

    def get_commit_datetime(self) -> str:
        if self._commit_datetime is not None:
            return self._commit_datetime
        commit_datetime = self._run_git_command(
            [
                "git",
                "log",
                "-1",
                "--format=%cd",
                "--date=format:%Y-%m-%d %H:%M:%S",
            ],
        )
        self._commit_datetime = commit_datetime.strip()
        return self._commit_datetime

    def get_branch(self) -> str:
        if self._branch is not None:
            return self._branch
        branch = self._run_git_command(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        )
        self._branch = branch.strip()
        return self._branch

    @staticmethod
    def _run_git_command(args: List[str]) -> str:
        result: str
        try:
            process_result = subprocess.run(
                args,
                cwd=os.getcwd(),
                capture_output=True,
                text=True,
                check=False,
            )
            if process_result.returncode != 0:
                return "N/A"
            result = process_result.stdout
        except subprocess.CalledProcessError:
            result = "N/A"
        except FileNotFoundError:
            result = "Git not available"
        return result



from typing import Optional

from strictdoc.backend.sdoc.models.model import SDocDocumentIF


class DocumentReference:
    def __init__(self) -> None:
        self._document: Optional[SDocDocumentIF] = None

    def get_document(self) -> Optional[SDocDocumentIF]:
        return self._document

    def set_document(self, document: SDocDocumentIF) -> None:
        self._document = document



"""
@relation(SDOC-SRS-72, scope=file)
"""

from typing import List

from reqif.parser import ReqIFParser, ReqIFZParser
from reqif.reqif_bundle import ReqIFBundle, ReqIFZBundle

from strictdoc.backend.reqif.p01_sdoc.reqif_to_sdoc_converter import (
    P01_ReqIFToSDocConverter,
)
from strictdoc.backend.sdoc.models.document import SDocDocument


class ReqIFReader:
    @staticmethod
    def read_from_file(input_path: str) -> List[SDocDocument]:
        converter = P01_ReqIFToSDocConverter()

        documents: List[SDocDocument]
        if input_path.endswith(".reqifz"):
            reqifz_bundle: ReqIFZBundle = ReqIFZParser.parse(input_path)
            assert len(reqifz_bundle.reqif_bundles) > 0
            documents = converter.convert_reqif_bundle(
                next(iter(reqifz_bundle.reqif_bundles.values())),
                enable_mid=True,
                import_markup="HTML",
            )
        else:
            reqif_bundle: ReqIFBundle = ReqIFParser.parse(input_path)
            documents = converter.convert_reqif_bundle(
                reqif_bundle,
                enable_mid=True,
                import_markup="HTML",
            )
        return documents



"""
@relation(SDOC-SRS-34, scope=file)
"""

from typing import Any, Optional

from strictdoc.helpers.auto_described import auto_described


@auto_described
class Req:
    def __init__(self, parent: Any, uid: str):
        assert isinstance(uid, str)
        assert len(uid) > 0

        self.parent = parent
        self.uid: str = uid

        self.ng_source_line: Optional[int] = None
        self.ng_source_column: Optional[int] = None



from typing import Any, Dict, Optional, Tuple, Type, Union

from strictdoc.core.graph.abstract_bucket import AbstractBucket


class OneToOneDictionary(AbstractBucket):
    def __init__(
        self,
        lhs_type: Type[Any],
        rhs_type: Union[Type[Any], Tuple[Type[Any], ...]],
    ):
        self._dict: Dict[Any, Any] = {}
        self._lhs_type: Type[Any] = lhs_type
        self._rhs_type: Union[Type[Any], Tuple[Type[Any], ...]] = rhs_type

    def has_link(self, *, lhs_node: Any) -> bool:
        assert isinstance(lhs_node, self._lhs_type), (lhs_node, self._lhs_type)
        return lhs_node in self._dict

    def get_count(self) -> int:
        return len(self._dict)

    def get_link_value(
        self, *, lhs_node: Any, edge: Optional[str] = None
    ) -> Any:
        assert edge is None
        assert isinstance(lhs_node, self._lhs_type)
        return self._dict[lhs_node]

    def get_link_value_weak(self, *, lhs_node: Any) -> Optional[Any]:
        assert isinstance(lhs_node, self._lhs_type)
        return self._dict.get(lhs_node, None)

    def create_link(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = None
    ) -> None:
        assert edge is None
        assert isinstance(lhs_node, self._lhs_type), (lhs_node, self._lhs_type)
        assert isinstance(rhs_node, self._rhs_type), (rhs_node, self._rhs_type)
        assert lhs_node not in self._dict, (
            f"OneToOneDictionary: Cannot create a link because lhs_node already exists: {lhs_node}."
        )
        self._dict[lhs_node] = rhs_node

    def delete_link(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = None
    ) -> None:
        assert edge is None
        assert isinstance(lhs_node, self._lhs_type), (lhs_node, self._lhs_type)
        assert isinstance(rhs_node, self._rhs_type), (rhs_node, self._rhs_type)
        assert lhs_node in self._dict
        del self._dict[lhs_node]



from typing import List, Union

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.core.transforms.validation_error import (
    MultipleValidationErrorAsList,
    SingleValidationError,
)
from strictdoc.helpers.cast import assert_cast


class DeleteRequirementCommand:
    def __init__(
        self,
        requirement: SDocNode,
        traceability_index: TraceabilityIndex,
    ):
        self.requirement: SDocNode = requirement
        self.traceability_index: TraceabilityIndex = traceability_index

    def validate(self) -> None:
        errors: List[str] = []
        document: SDocDocument = assert_cast(
            self.requirement.get_document(), SDocDocument
        )
        document_iterator = DocumentCachingIterator(document=document)
        for document_node_, _ in document_iterator.all_node_content(
            self.requirement,
            print_fragments=True,
        ):
            if not isinstance(document_node_, SDocNode):
                continue
            nodes_with_incoming_links = [
                document_node_
            ] + document_node_.get_anchors()
            for node_ in nodes_with_incoming_links:
                try:
                    self.traceability_index.validate_can_remove_node(node=node_)
                except SingleValidationError as exception_:
                    errors.append(exception_.args[0])
        if len(errors) > 0:
            raise MultipleValidationErrorAsList("NOT_RELEVANT", errors)

    def perform(self) -> None:
        self.validate()

        self.traceability_index.delete_requirement(self.requirement)

        requirement_parent: Union[SDocSectionIF, SDocDocumentIF, SDocNodeIF] = (
            self.requirement.parent
        )

        requirement_parent.section_contents.remove(self.requirement)

        self.traceability_index.update_last_updated()



from collections import defaultdict
from typing import Dict, List, Optional, Union

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.model import (
    SDocDocumentIF,
    SDocNodeIF,
    SDocSectionIF,
)
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import (
    TraceabilityIndex,
)
from strictdoc.core.transforms.constants import NodeCreationOrder
from strictdoc.core.transforms.validation_error import (
    MultipleValidationError,
    SingleValidationError,
)
from strictdoc.export.html.form_objects.section_form_object import (
    SectionFormObject,
)
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.mid import MID


class UpdateSectionCommand:
    def __init__(
        self,
        form_object: SectionFormObject,
        section: SDocSection,
        traceability_index: TraceabilityIndex,
    ) -> None:
        self.form_object: SectionFormObject = form_object
        self.section: SDocSection = section
        self.traceability_index: TraceabilityIndex = traceability_index

    def perform(self) -> None:
        errors: Dict[str, List[str]] = defaultdict(list)
        validation_error = MultipleValidationError(
            "Section form has not passed validation.", errors
        )

        form_object = self.form_object
        section = self.section
        traceability_index = self.traceability_index

        if len(form_object.section_title) == 0:
            errors["section_title"].append("Section title must not be empty.")

        # Validate UID.
        if section.reserved_uid is not None:
            # This is case where an existing section UID is being removed.
            # We have to check if this UID has incoming links to it, and if so,
            # raise a validation error.
            if len(form_object.section_uid) == 0:
                try:
                    traceability_index.validate_section_can_remove_uid(
                        section=section
                    )
                except SingleValidationError as validation_error_:
                    errors["section_uid"].append(validation_error_.args[0])
            else:
                if section.reserved_uid != form_object.section_uid:
                    section_incoming_links: Optional[List[InlineLink]] = (
                        traceability_index.get_incoming_links(section)
                    )
                    if (
                        section_incoming_links is not None
                        and len(section_incoming_links) > 0
                    ):
                        errors["section_uid"].append(
                            "Renaming a section UID when a section has "
                            "incoming links is not supported yet. "
                            "For now, please use a command-line process to "
                            "rename the section UID and all links that refer "
                            "to it."
                        )
        if len(form_object.section_uid) > 0:
            try:
                traceability_index.validate_can_create_uid(
                    form_object.section_uid, section.reserved_mid
                )
            except SingleValidationError as validation_error_:
                errors["section_uid"].append(validation_error_.args[0])

        if len(errors) > 0:
            raise validation_error

        # Updating section title.
        if (
            form_object.section_title is not None
            and len(form_object.section_title) > 0
        ):
            section.title = form_object.section_title
        else:
            raise AssertionError(
                f"Should not reach here: {form_object.section_title}"
            )

        # We have passed the validations if we reach this point, so we can
        # just assume we can safely delete the previous section UID
        # associations.
        traceability_index.delete_section(section)

        # Updating section UID.
        if len(form_object.section_uid) > 0:
            section_uid = assert_cast(form_object.section_uid, str)
            section.reserved_uid = section_uid
        else:
            section.reserved_uid = None

        traceability_index.create_section(section)


class CreateSectionCommand:
    def __init__(
        self,
        form_object: SectionFormObject,
        whereto: str,
        reference_mid: str,
        traceability_index: TraceabilityIndex,
        config: ProjectConfig,
    ):
        self.form_object: SectionFormObject = form_object
        self.whereto: str = whereto
        self.reference_mid: str = reference_mid
        self.traceability_index: TraceabilityIndex = traceability_index
        self.config: ProjectConfig = config

        self._created_section: Optional[SDocSection] = None

    def get_created_section(self) -> SDocSection:
        assert isinstance(self._created_section, SDocSection)
        return self._created_section

    def perform(self) -> None:
        errors: Dict[str, List[str]] = defaultdict(list)
        validation_error = MultipleValidationError(
            "Section form has not passed validation.", errors
        )

        form_object = self.form_object
        whereto = self.whereto
        reference_mid = self.reference_mid
        traceability_index = self.traceability_index

        reference_node: Union[SDocDocumentIF, SDocSectionIF, SDocNodeIF] = (
            assert_cast(
                traceability_index.get_node_by_mid(MID(reference_mid)),
                (SDocDocumentIF, SDocSectionIF, SDocNodeIF),
            )
        )

        document: SDocDocumentIF  # type: ignore[unreachable]
        if isinstance(reference_node, SDocDocument):
            if not reference_node.document_is_included():
                document = reference_node
            else:
                # Edge case: a reference node is a root one from included document.
                # Only when a child node is added, the document is set to the included document.
                # Otherwise, to the including document.
                if whereto == "child":
                    document = reference_node
                else:
                    document = assert_cast(
                        reference_node.get_including_document(), SDocDocument
                    )
        elif document_ := reference_node.get_document():
            assert isinstance(document_, SDocDocument)
            document = document_
        else:
            raise AssertionError(
                f"Unexpected error: Expected a document node to be available for reference node: {reference_node}."
            )

        if len(form_object.section_title) == 0:
            errors["section_title"].append("Section title must not be empty.")

        if len(
            form_object.section_uid
        ) > 0 and traceability_index.has_node_connections(
            form_object.section_uid
        ):
            existing_section = assert_cast(
                traceability_index.get_node_by_uid(form_object.section_uid),
                SDocSection,
            )
            errors["section_uid"].append(
                f"The chosen UID must be unique. "
                f"There is another section '{existing_section.title}' with "
                f"a UID '{form_object.section_uid}'."
            )

        if len(errors) > 0:
            raise validation_error

        parent: Union[SDocDocumentIF, SDocSectionIF]
        if whereto == NodeCreationOrder.CHILD:
            parent = reference_node
            insert_to_idx = len(parent.section_contents)
        elif whereto == NodeCreationOrder.BEFORE:
            assert isinstance(
                reference_node, (SDocDocument, SDocNode, SDocSection)
            ), reference_node
            # Be aware of an edge case besides all normal cases:
            # A reference node can be a root node of an included document.
            if not isinstance(reference_node, SDocDocument):
                parent = reference_node.parent
                insert_to_idx = parent.section_contents.index(reference_node)
            else:
                parent = assert_cast(
                    reference_node.ng_including_document_from_file.parent,
                    (SDocDocumentIF, SDocSectionIF),
                )
                assert (
                    reference_node.ng_including_document_from_file is not None
                )
                insert_to_idx = parent.section_contents.index(
                    reference_node.ng_including_document_from_file
                )
        elif whereto == NodeCreationOrder.AFTER:
            assert isinstance(
                reference_node, (SDocDocument, SDocNode, SDocSection)
            )
            if isinstance(reference_node, SDocDocument):
                assert reference_node.document_is_included()
                assert (
                    reference_node.ng_including_document_from_file is not None
                )
                parent = reference_node.ng_including_document_from_file.parent
                insert_to_idx = (
                    parent.section_contents.index(
                        reference_node.ng_including_document_from_file
                    )
                    + 1
                )
            else:
                parent = reference_node.parent
                insert_to_idx = (
                    parent.section_contents.index(reference_node) + 1
                )
        else:
            raise NotImplementedError

        section = SDocSection(
            parent=parent,
            mid=None,
            uid=None,
            custom_level=None,
            title="NOT_RELEVANT",
            requirement_prefix=None,
            section_contents=[],
        )

        if len(form_object.section_uid) > 0:
            section_uid = assert_cast(form_object.section_uid, str)
            section.uid = section_uid
            section.reserved_uid = section_uid

        section.node_id = MID(form_object.section_mid)
        section.ng_document_reference = DocumentReference()
        section.ng_document_reference.set_document(document)
        section.ng_including_document_reference = (
            document.ng_including_document_reference
        )

        parent.section_contents.insert(insert_to_idx, section)

        traceability_index.create_section(section)

        # Updating section title.
        if (
            form_object.section_title is not None
            and len(form_object.section_title) > 0
        ):
            section.title = form_object.section_title

        self._created_section = section
